-- -----------------------------------------------------------------------------
-- -----------------------------------------------------------------------------
--  Hammer of Justas
--
-- Copyright (C) 2010-2020
-- By Jonathan E. Wright
--
-- FILE: HoJ_Paladin.lua
-- DESC: Paladin-specific functions.

--[[

Shadowlands TODO:
- Ret
  - Rotation
    X incorporate consecration
    X consecration timer not working correctly -- probably because it doesn't buff ret paladin
    - Seraphim tracking
    - Test Justicar's Vengeance
    - Test inquisition
    X Don't suggest Hammer of Wrath during Avenging Wrath unless HoW is rank 2 or higher
    X Empyrean Power is not working
  - Cooldowns
    - Holy Avenger 
  - if a party or raid member is about to die, suggest retribution aura
- Prot
  - uses holy power now!
  - GetSpellCharges() for Shield of the Righteous -- not necessary?
--]]

local addonName, HoJ = ...

-- DEFAULT_CHAT_FRAME:AddMessage( "HoJ_Paladin.lua" )

local paladinSpellInfo = HoJ._paladinSpellInfos

local TEMPLARS_VERDICT_ID = 224266
local RIGHTEOUS_VERDICT_ID = 267611
local JUDGMENT_ID = 20271

-- -----------------------------------------------------------------------------
-- HoJ:Paladin_DamageEventHandler
-- -----------------------------------------------------------------------------
function HoJ:Paladin_DamageEventHandler( spellId, srcGUID, destGUID, playerIsSource, playerIsTarget, isSpell, actualDamageDone )
	-- HoJ:Log( "Paladin_DamageEvenHandler" )

	local curTime = GetTime()
	local scs = HoJ._sessionCombatStats
	
    -- we have to track the judgment debuff ourselves because it will have been consumed by
	-- the TV attack by the time we see the log message that it hit.
	if ( spellId == paladinSpellInfo.judgment.id and playerIsSource ) then
		local js = scs._judgment
		js._expireTime = curTime + 15.0
		js._targetUnitGUID = destGUID
		-- HoJ:Log( "Added judgment debuff!" )
	end

	if ( spellId == paladinSpellInfo.templarsVerdictDamage.id and playerIsSource ) then
		-- HoJ:Log( "TV!" )

		scs._numTemplarsVerdict = scs._numTemplarsVerdict + 1

		local js = scs._judgment 
		if ( js._targetUnitGUID == destGUID and js._expireTime > curTime ) then
			js._targetUnitGUID = 0
			js._expireTime = 0.0

			scs._tvsWithJudgement = scs._tvsWithJudgement + 1

			-- HoJ:Log( "TV with Judgment!" )
		end

		local rvs = scs._righteousVerdict
		if ( rvs._expireTime > curTime ) then
			-- HoJ:Log( "TV with Righteous Verdict!" )
			scs._tvsWithRighteousVerdict = scs._tvsWithRighteousVerdict + 1
		end

		-- update RV after we've checked to see if there was already one when we cast TV
		-- we have to track this when it's applied vs. looking up player buffs because
		-- by the time we get this message, the RV buff will from this attack will show
		-- up on the player's buffs, when we want to know if there was an RV buff on the
		-- player prior to the attack.
		if ( HoJ:IsTalentSelected( paladinSpellInfo.righteousVerdict ) ) then
			-- HoJ:Log( "Righteous Verdict buff applied!" )
			rvs._expireTime = curTime + 6.0
			rvs._targetUnitGUID = srcGUID
		end
    end
end

-- -----------------------------------------------------------------------------
-- HoJ:Paladin_ReportSessionCombatStats
-- -----------------------------------------------------------------------------
function HoJ:Paladin_ReportSessionCombatStats( stats )
	HoJ:Log( "Session " .. stats._sessionId .. " combat stats:" )
	HoJ:Log( format( "  TVs: %i", stats._numTemplarsVerdict ) )
	if ( stats._numTemplarsVerdict > 0 ) then
		HoJ:Log( format( "  TVs with Judgment: %i (%.1f%%)", stats._tvsWithJudgement, 
				( stats._tvsWithJudgement / stats._numTemplarsVerdict ) * 100.0 ) )
		HoJ:Log( format( "  TVs with Righteous Verdict: %i (%.1f%%)", stats._tvsWithRighteousVerdict,
				( stats._tvsWithRighteousVerdict / stats._numTemplarsVerdict ) * 100.0 ) )
	end
end

-- -----------------------------------------------------------------------------
-- HoJ:DebugRule_LayOnhHands
-- -----------------------------------------------------------------------------
function HoJ:DebugRule_LayOnHands( rule )
    local name = rule.spellInfo.name
    if ( rule._debugName ~= nil ) then
        name = rule._debugName
    end

    HoJ:Log( "=== " .. name .. " success! ===" )
    HoJ:Log( format( "Player Health = %.0f", HoJ._playerHealth ) )
    HoJ:Log( format( "Player TTL = %.2f", HoJ._playerTimeToLive ) )
    HoJ:Log( format( "NIDPS = %.2f", HoJ._netIncomingDps ) )
    HoJ:Log( format( "AIDPS = %.2f", HoJ._averageIncomingDps ) )
    HoJ:Log( format( "Player HR = %.2f", HoJ._playerHealthRatio ) )

    local scale = 5.0
    local scaledDps = ( HoJ._netIncomingDps * scale )
    local dpschr = scaledDps / HoJ._playerHealth

    HoJ:Log( format( "NIDPS * %.2f = %.2f", scale, scaledDps ) )
    HoJ:Log( format( "NIDPSCHR = %.2f", dpschr ) )
end

-- -----------------------------------------------------------------------------
-- HoJ:Paladin_SelectSingleOrMultitarget
-- returns
--   rotationRules
--   interruptRules
--   mitigationRules
--   dpsCooldownRules
-- -----------------------------------------------------------------------------
function HoJ:Paladin_SelectSingleOrMultitarget( curSpecInfo, numAttackers, threshold )
    if ( self._numMeleeAttackers > threshold ) then
        return curSpecInfo._rotationRules._multiTarget, 
                curSpecInfo._interruptRules._multiTarget,
                curSpecInfo._mitigationRules._multiTarget,
                curSpecInfo._dpsCooldownRules._multiTarget
    else
        return curSpecInfo._rotationRules._singleTarget, 
                curSpecInfo._interruptRules._singleTarget,
                curSpecInfo._mitigationRules._singleTarget,
                curSpecInfo._dpsCooldownRules._singleTarget
    end
end

-- -----------------------------------------------------------------------------
-- HoJ:Paladin_CheckBlessings
-- CLASSFIX - paladin specific
-- -----------------------------------------------------------------------------
function HoJ:Paladin_CheckBlessings( activeBlessingsList )
	if ( not HoJ._showBuffsFrame ) then
		return
	end

	-- this has to be called every time we check blessings to update non-party member blessings
	if ( self._updateFrameNum >= self._nextUpdateBlessingFrame ) then
		HoJ:UpdateBlessings( activeBlessingsList )
	end

	--HoJ:DebugMsg( "CheckBlessings )

	-- dont show any blessing buffs for a second after entering world
	if ( self._updateFrameNum < self._enteredWorldUpdateFrameNum + HoJ.UPDATES_PER_SEC ) then
		return
	end

    -- SLREMOVE
--[[
	if ( HoJ._level >= HoJ.SpellInfo.blessingOfKings.minLevel ) then
		-- handle party sync disabling this buff
		if ( not C_SpellBook.IsSpellDisabled( HoJ.SpellInfo.blessingOfKings.id ) ) then
			local kingsBlessing = activeBlessingsList[HoJ.SpellInfo.blessingOfKings.id]
			if ( kingsBlessing.start <= 0.0 ) then
				--HoJ:DebugMsg( "Adding Blessing of Kings icon" )
				HoJ:AddIcon( HoJ._buffs, HoJ.SpellInfo.blessingOfKings )
			end
		end
	end

	if ( HoJ._level >= HoJ.SpellInfo.blessingOfWisdom.minLevel ) then
		-- handle party sync disabling this buff
		if ( not C_SpellBook.IsSpellDisabled( HoJ.SpellInfo.blessingOfWisdom.id ) ) then
			local wisdomBlessing = activeBlessingsList[HoJ.SpellInfo.blessingOfWisdom.id]
			if ( wisdomBlessing.start <= 0.0 ) then
				--HoJ:DebugMsg( "Adding Blessing of Wisdom icon" )
				HoJ:AddIcon( HoJ._buffs, HoJ.SpellInfo.blessingOfWisdom )
			end
		end
    end
--]]
end

-- -----------------------------------------------------------------------------
-- HoJ:Paladin_Update
-- -----------------------------------------------------------------------------
function HoJ:Paladin_Update( curSpecInfo, updateCooldowns, updateBuffs, updateRotation )
    -- HoJ:DebugMsg( "Updating spec " .. specInfo._name );

    local rotationRules, interruptRules, mitigationRules, dpsCooldownRules = curSpecInfo._selectSingleOrMultiTargetFn( 
        HoJ, curSpecInfo, 
        self._numMeleeAttackers, 2 )

    -- update buffs
    if ( updateBuffs ) then
        local startBuffsMs = HoJ.ProfileBegin2( "Buffs" )

        HoJ:ClearIconList( HoJ._buffs );

        HoJ:CheckProcdBuffs( curSpecInfo._activeBuffList, curSpecInfo._procdBuffList );
        if ( curSpecInfo._checkBlessingsFn ~= nil ) then
            curSpecInfo._checkBlessingsFn( self, curSpecInfo._activeBlessings )
        end

        HoJ.ProfileFunction( "UpdateIcons2:Buffs:", HoJ.UpdateIcons2, self, HoJ._buffs, 
                curSpecInfo._buffSortPriorities, HoJ._buffFrames, true )

        HoJ.ProfileEnd( "Buffs", startBuffsMs )
    end

    -- update cooldowns
    if ( updateCooldowns ) then
        local startCooldownsMs = HoJ.ProfileBegin2( "Cooldowns" )

        HoJ.ProfileFunction( "ClearIconList", HoJ.ClearIconList, self, HoJ._cooldowns )

        HoJ.ProfileFunction( "CheckRules(dpsCooldowns)", HoJ.CheckRules, self, self._cooldowns, 
                self._showCooldownsFrame, dpsCooldownRules, true, true )
        HoJ.ProfileFunction( "CheckRules(interrupts)", HoJ.CheckRules, self, self._cooldowns, 
                self._showCooldownsFrame, interruptRules, true, true )
        HoJ.ProfileFunction( "CheckRules(mitigations)", HoJ.CheckRules, self, self._cooldowns, 
                self._showCooldownsFrame, mitigationRules, true, true )

        HoJ.ProfileFunction( "UpdateIcons2:Cooldowns:", HoJ.UpdateIcons2, self, HoJ._cooldowns, 
                curSpecInfo._cooldownSortPriorities, HoJ._cooldownFrames, true )

        HoJ.ProfileEnd( "Cooldowns", startCooldownsMs )
    end

    -- update rotation
    if ( updateRotation ) then
        local startRotationMs = HoJ.ProfileBegin2( "Rotation" )

        HoJ:ClearIconList( HoJ._rotationSkills )

        HoJ:CheckDPSRotation( HoJ._rotationSkills, rotationRules )

        HoJ.ProfileFunction( "UpdateIcons2:Rotation:", HoJ.UpdateIcons2, self, HoJ._rotationSkills, 
                { }, HoJ._rotationFrames, false );

        HoJ.ProfileEnd( "Rotation", startRotationMs )
    end
end

-- -----------------------------------------------------------------------------
-- HoJ:Paladin_Update_Holy
-- -----------------------------------------------------------------------------
function HoJ:Paladin_Update_Holy( specInfo, updateCooldowns, updateBuffs, updateRotation )
    -- HoJ:DebugMsg( "Updating spec " .. specInfo._name );
    if ( not specInfo._enabled ) then
        HoJ:DebugMsg( "Spec " .. specInfo._name .. " is disabled!" )
        return
    end    
end

-- -----------------------------------------------------------------------------
-- HoJ:SetHPAlphas
-- -----------------------------------------------------------------------------
function HoJ:SetHPAlphas( alpha, numVisible )
    if ( not HoJ._showHolyPowerFrame and not HoJ._showTemplarsVerdictFrame ) then
        return
    end

    if ( alpha < 0.0 ) then
		alpha = 0.0
	elseif ( alpha > 1.0 ) then
		alpha = 1.0
    end
    
	local holyPowerAlpha = alpha * self._holyPowerAlpha * self._globalAlpha
	local templarsVerdictAlpha = alpha * self._holyPowerAlpha * self._globalAlpha

    -- this should never happen again with the fix to wait for _needsRest before rendering
    if ( numVisible < 0 or numVisible > 5 ) then
        HoJ:Error( "numVisible = " .. numVisible .. "!" )
        numVisible = 0
    end

	for i = 1, numVisible do
		--HoJ:DebugMsg( "Setting alpha " .. i );
		HoJ._HoPo[i]:SetAlpha( holyPowerAlpha );
		HoJ._HoPoGlow[i]:SetAlpha( holyPowerAlpha );
	end

	for i = numVisible + 1, 5 do
		--HoJ:DebugMsg( "Clearing alpha " .. i );
		HoJ._HoPo[i]:SetAlpha( 0.0 );
		HoJ._HoPoGlow[i]:SetAlpha( 0.0 );
	end

	-- set alpha for big graphic
	if ( numVisible < 3 ) then
		HoJFrame.HPMaxed:SetAlpha( 0.0 );
		HoJFrame.HPMaxedGlow:SetAlpha( 0.0 );
	else
		HoJFrame.HPMaxed:SetAlpha( templarsVerdictAlpha );
		HoJFrame.HPMaxedGlow:SetAlpha( templarsVerdictAlpha );
	end
end

-- -----------------------------------------------------------------------------
-- HoJ:HideHolyPowerFrames
-- -----------------------------------------------------------------------------
function HoJ:HideHolyPowerFrames()
    -- HoJ:Log( "HideHolyPowerFrames" )
    -- hide paladin-specific UI for Holy Power
    for i, v in ipairs( HoJ._HoPo ) do
        --HoJ:DebugMsg( "Setting alpha " .. i );
        v:SetAlpha( 0.0 );
        HoJ._HoPoGlow[i]:SetAlpha( 0.0 )
    end

    -- hide Templar's Verdict frame
    HoJFrame.HPMaxed:SetAlpha( 0.0 );
    HoJFrame.HPMaxedGlow:SetAlpha( 0.0 );
end

-- -----------------------------------------------------------------------------
-- HoJ:Paladin_Reset
-- -----------------------------------------------------------------------------
function HoJ:Paladin_Reset()
    -- HoJ:Log( "Paladin_Reset" )
    self._glowAlpha = 0.0
    self:HideHolyPowerFrames()
end

-- -----------------------------------------------------------------------------
-- HoJ:UpdateGlow
-- -----------------------------------------------------------------------------
function HoJ:UpdateGlow( frame, elapsed )
	if ( self._glowAlpha > self._maxGlowAlpha ) then
		self._glowScale = 1.0 - ( 4.0 * elapsed )
	elseif ( self._glowAlpha < self._minGlowAlpha ) then
		self._glowScale = 1.0 + ( 4.0 * elapsed )
	end

	self._glowAlpha = self._glowAlpha * self._glowScale
	if ( self._glowAlpha <= 0.0 ) then
		self._glowAlpha = self._minGlowAlpha
	elseif ( self._glowAlpha > 1.0 ) then
		self._glowAlpha = self._maxGlowAlpha
	end
	--self:DebugMsg( "_glowAlpha = " .. self._glowAlpha )
	--self:DebugMsg( "_glowScale = " .. self._glowScale )

	finalAlpha = self._glowAlpha
	if ( IsMounted() ) then
		finalAlpha = finalAlpha * 0.1875
	end

	self:SetHPAlphas( finalAlpha, self._power )

	--HoJ_RotationFrame:SetAlpha( 0.666 )
end

-- -----------------------------------------------------------------------------
-- HoJ:Paladin_UpdateRender
-- -----------------------------------------------------------------------------
function HoJ:Paladin_UpdateRender( frame, elapsed )
    -- HoJ:DebugMsg( "Paladin_UpdateRender" )

    local curSpecInfo = self._curSpecInfo
    if ( curSpecInfo._powerType == 9 ) then
        HoJ:UpdateGlow( frame, elapsed )
    else
        HoJ:HideHolyPowerFrames()
    end
end

local holyMeleeAoESpells = {
}

local protMeleeAoESpells = {
    [81297] = { -- consecration damage
        _id = 81297,
        _name = HoJ:GetSpellName( 81297 )
    }
}

local retMeleeAoESpells = {
    [53385] = { -- divine storm damage
        _id = 53385,
        _name = HoJ:GetSpellName( 53385 )
    }        
}

local lowHealthMitigationList = {
    { spellInfo = paladinSpellInfo.handOfTheProtector },
    { spellInfo = paladinSpellInfo.lightOfTheProtector, replacedByTalent = paladinSpellInfo.handOfTheProtector }
}

-- MODERATE MITIGATION SKILL PRIORITY
local moderateMitigationList = {
    { spellInfo = paladinSpellInfo.shieldOfTheRighteous },
    { spellInfo = paladinSpellInfo.giftOfTheNaaru },
    { spellInfo = paladinSpellInfo.handOfTheProtector },
    { spellInfo = paladinSpellInfo.lightOfTheProtector, replacedByTalent = paladinSpellInfo.handOfTheProtector },
    { spellInfo = paladinSpellInfo.ardentDefender },
    { spellInfo = paladinSpellInfo.guardianOfAncientKings }
}

-- EMERGENCY MITIGATION SKILL PRIORITY
-- in situations where we're about to die very soon, prefer to
--
-- 1) heal using Hand / Light of the Protector
-- 2) heal using Lay on Hands
-- 3) cast ardent defender (because this will prevent a death and heal to 20%)
-- 4) cast guardian of ancient kings
-- 4) blessing of protection (assuming incoming damage is physical)
-- 5) divine shield (last resort, should be followed by a taunt-and-cancel if tanking)
--
-- not really sure about these priorities... in some situations where I'm pretty sure
-- I'm not going to die I prefer to to use Ardent Defender every two minutes as a
-- damage-reducer only.
local emergencyMitigationList = {
    { spellInfo = paladinSpellInfo.layOnHands, excludedByBuffId = paladinSpellInfo.forbearance.id },
    { spellInfo = paladinSpellInfo.handOfTheProtector },
    { spellInfo = paladinSpellInfo.lightOfTheProtector, replacedByTalent = paladinSpellInfo.handOfTheProtector },
    { spellInfo = paladinSpellInfo.ardentDefender },
    { spellInfo = paladinSpellInfo.guardianOfAncientKings },
    { spellInfo = paladinSpellInfo.blessingOfProtection },
    { spellInfo = paladinSpellInfo.divineShield, excludedByBuffId = paladinSpellInfo.forbearance.id }
}

--[[
test = {
    aType	-- type of the first argument
    aConst	-- constant value of the first argument
    bType	-- type of the second argument
    bConst	-- constant value of the second argument
    op 		-- the operator to use ( a op b )
}
]]

--------------------------------------------------------------------------------
-- RETRIBUTION
--------------------------------------------------------------------------------

local retRotationRules = {
    {
        spellInfo = paladinSpellInfo.inquisition,
        -- _debugName = "inquisition",
        tests = {
            { aType = HoJ.TYPE_POWER, aConst = 0, bType = HoJ.TYPE_CONSTANT, bConst = 2, op = HoJ.OP_GREATER_EQUAL },
            { aType = HoJ.TYPE_PLAYER_BUFF_REMAINING, aConst = paladinSpellInfo.inquisition.id, bType = HoJ.TYPE_CONSTANT, bConst = 2.0, op = HoJ.OP_LESS_EQUAL },
            { aType = HoJ.TYPE_RESULT, aConst = 1, bType = HoJ.TYPE_RESULT, bConst = 2, op = HoJ.OP_AND },
        }
    },
    {
        spellInfo = paladinSpellInfo.justicarsVengeance,
        -- _debugName = "justicarsVengeance",
        tests = {
            { aType = HoJ.TYPE_POWER, aConst = 0, bType = HoJ.TYPE_CONSTANT, bConst = 5, op = HoJ.OP_GREATER_EQUAL },
            { aType = HoJ.TYPE_TARGET_RANGE, aConst = paladinSpellInfo.justicarsVengeance.range, bType = HoJ.TYPE_CONSTANT, bConst = true, op = HoJ.OP_EQUAL },
            { aType = HoJ.TYPE_RESULT, aConst = 1, bType = HoJ.TYPE_RESULT, bConst = 2, op = HoJ.OP_AND },
            { aType = HoJ.TYPE_PLAYER_HEALTH_RATIO, aConst = 0, bType = HoJ.TYPE_MAJOR_HEAL_HEALTH_RATIO, bConst = 0, op = HoJ.OP_LESS_EQUAL },
            { aType = HoJ.TYPE_RESULT, aConst = 3, bType = HoJ.TYPE_RESULT, bConst = 4, op = HoJ.OP_AND }

        }
    },
    {
        spellInfo = paladinSpellInfo.consecration,
        -- _debugName = "consecration",
        tests = {
            { aType = HoJ.TYPE_TARGET_RANGE, aConst = paladinSpellInfo.consecration.range, bType = HoJ.TYPE_CONSTANT, bConst = true, op = HoJ.OP_EQUAL },
            { aType = HoJ.TYPE_PLAYER_BUFF_REMAINING, aConst = paladinSpellInfo.consecration.id, bType = HoJ.TYPE_CONSTANT, bConst = 1.5, op = HoJ.OP_LESS_EQUAL },
            { aType = HoJ.TYPE_RESULT, aConst = 1, bType = HoJ.TYPE_RESULT, bConst = 2, op = HoJ.OP_AND },
        }
    },
    {
        spellInfo = paladinSpellInfo.executionSentence,
        -- _debugName = "executionSentence",
        -- _debugFail = true,
        tests = {
            { aType = HoJ.TYPE_POWER, aConst = 0, bType = HoJ.TYPE_CONSTANT, bConst = 3, op = HoJ.OP_GREATER_EQUAL },
            { aType = HoJ.TYPE_TARGET_RANGE, aConst = paladinSpellInfo.executionSentence.range, bType = HoJ.TYPE_CONSTANT, bConst = true, op = HoJ.OP_EQUAL },
            { aType = HoJ.TYPE_RESULT, aConst = 1, bType = HoJ.TYPE_RESULT, bConst = 2, op = HoJ.OP_AND }
        }
    },
    -- if we have Divine Purpose up and we're in range of TV
    -- if (
    --   DP it's about to run out
    --   OR if we have less than two Crusader Strike charges
    --   OR if Judgment is on cooldown for more than ~GCD
    --   OR if Blade of Justice is on cooldown for more than ~GCD
    -- )
    -- USE TV 
    {
        spellInfo = paladinSpellInfo.templarsVerdict,
        -- _debugName = "TV",
        tests = {
            { aType = HoJ.TYPE_PLAYER_BUFF_ACTIVE, aConst = paladinSpellInfo.divinePurposeBuff.id, bType = HoJ.TYPE_CONSTANT, bConst = true, op = HoJ.OP_EQUAL },
            { aType = HoJ.TYPE_TARGET_RANGE, aConst = paladinSpellInfo.templarsVerdict.range, bType = HoJ.TYPE_CONSTANT, bConst = true, op = HoJ.OP_EQUAL },
            { aType = HoJ.TYPE_RESULT, aConst = 1, bType = HoJ.TYPE_RESULT, bConst = 2, op = HoJ.OP_AND },
            { aType = HoJ.TYPE_PLAYER_BUFF_REMAINING, aConst = paladinSpellInfo.divinePurposeBuff.id, bType = HoJ.TYPE_CONSTANT, bConst = { 0.0, 2.0 }, op = HoJ.OP_IN_RANGE_MIN_EX_MAX_IN },
            { aType = HoJ.TYPE_SKILL_CHARGES, aConst = paladinSpellInfo.crusaderStrike, bType = HoJ.TYPE_CONSTANT, bConst = 2, op = HoJ.OP_LESS },
            { aType = HoJ.TYPE_RESULT, aConst = 4, bType = HoJ.TYPE_RESULT, bConst = 5, op = HoJ.OP_OR },
            { aType = HoJ.TYPE_SKILL_COOLDOWN_REMAINING, aConst = paladinSpellInfo.judgment, bType = HoJ.TYPE_CONSTANT, bConst = 2, op = HoJ.OP_GREATER_EQUAL },
            { aType = HoJ.TYPE_SKILL_COOLDOWN_REMAINING, aConst = paladinSpellInfo.bladeOfJustice, bType = HoJ.TYPE_CONSTANT, bConst = 2, op = HoJ.OP_GREATER_EQUAL },
            { aType = HoJ.TYPE_RESULT, aConst = 7, bType = HoJ.TYPE_RESULT, bConst = 8, op = HoJ.OP_OR },
            { aType = HoJ.TYPE_RESULT, aConst = 6, bType = HoJ.TYPE_RESULT, bConst = 9, op = HoJ.OP_OR },
            { aType = HoJ.TYPE_RESULT, aConst = 3, bType = HoJ.TYPE_RESULT, bConst = 10, op = HoJ.OP_AND },
        }
    },
    {
        spellInfo = paladinSpellInfo.wakeOfAshes,
        -- _debugName = "wakeOfAshes",
        tests = {
            { aType = HoJ.TYPE_POWER, aConst = 0, bType = HoJ.TYPE_CONSTANT, bConst = 1, op = HoJ.OP_LESS_EQUAL },
            { aType = HoJ.TYPE_TARGET_RANGE, aConst = paladinSpellInfo.wakeOfAshes.range, bType = HoJ.TYPE_CONSTANT, bConst = true, op = HoJ.OP_EQUAL },
            { aType = HoJ.TYPE_RESULT, aConst = 1, bType = HoJ.TYPE_RESULT, bConst = 2, op = HoJ.OP_AND }
        }
    },
    {
        -- empyrean power if it is about to expire
        spellInfo = paladinSpellInfo.empyreanPower,
        -- _debugName = "empyreanPower_expiring",
        tests = {
            { aType = HoJ.TYPE_PLAYER_BUFF_ACTIVE, aConst = paladinSpellInfo.empyreanPowerBuff.id, bType = HoJ.TYPE_CONSTANT, bConst = true, op = HoJ.OP_EQUAL },
            { aType = HoJ.TYPE_TARGET_RANGE, aConst = paladinSpellInfo.empyreanPower.range, bType = HoJ.TYPE_CONSTANT, bConst = true, op = HoJ.OP_EQUAL },
            { aType = HoJ.TYPE_RESULT, aConst = 1, bType = HoJ.TYPE_RESULT, bConst = 2, op = HoJ.OP_AND },
            { aType = HoJ.TYPE_PLAYER_BUFF_REMAINING, aConst = paladinSpellInfo.empyreanPowerBuff.id, bType = HoJ.TYPE_CONSTANT, bConst = 2, op = HoJ.OP_LESS_EQUAL },
            { aType = HoJ.TYPE_RESULT, aConst = 3, bType = HoJ.TYPE_RESULT, bConst = 4, op = HoJ.OP_AND },
        }
    },
    {
        spellInfo = paladinSpellInfo.hammerOfWrath,
        -- _debugName = "hammerOfWrath",
        -- _debugSuccess = true,
        tests = {
            { aType = HoJ.TYPE_POWER, aConst = 0, bType = HoJ.TYPE_CONSTANT, bConst = 4, op = HoJ.OP_LESS_EQUAL },
            { aType = HoJ.TYPE_TARGET_RANGE, aConst = paladinSpellInfo.hammerOfWrath.range, bType = HoJ.TYPE_CONSTANT, bConst = true, op = HoJ.OP_EQUAL },
            { aType = HoJ.TYPE_RESULT, aConst = 1, bType = HoJ.TYPE_RESULT, bConst = 2, op = HoJ.OP_AND },
            { aType = HoJ.TYPE_TARGET_HEALTH_RATIO, aConst = 0, bType = HoJ.TYPE_CONSTANT, bConst = 0.2, op = HoJ.OP_LESS_EQUAL },
            { aType = HoJ.TYPE_RESULT, aConst = 3, bType = HoJ.TYPE_RESULT, bConst = 4, op = HoJ.OP_AND },

        }
    },
    -- NOTE: instead of wasting the 1 holy power to cast judgment at 5, this has been changed to only cast judgement before TV at 4 holy power.
    -- If judgment is off cooldown, we have 4 or 5 holy power, we're in judgment range, and we don't have judgment on our target,
    -- cast judgment so we get the 25% damage bonus when we use TV next.
    -- If already at 5 HoPo, this wastes 1 HoPo to get the 25% increase and its use is debatable -- I'm not sure what that 1 wasted
    -- HoPo amounts to in DPS.
    -- At 4 HoPo, this immediately applies Judgment to bring us to 5 HoPo where the next skill will be TV/DS
    -- Without this rule and TV == 4 we would normally skip TV since HoPo < 5, but use HoW since it will give us 4 + 1 == 5 TV
    -- If judgment is off cooldown, then that wastes an opportunity to get the 25% damage bonus on the next HoPo spender
    -- We could alternatively modify the HoW test to not fire if Judgment is off CD, or just prioritize Judgment over HoW, which some sites do.
    -- It's tempting to just move HoW priority after BoJ and Judgment, but it does have a higher overall DPS when it's up.
    -- If were were selecting this manually it would be much harder to coordinate when to prioritize a 4-HoPo Judgment vs a 4-HoPo HoW
    -- so it might still make sense to just have HoW the higher priority.
    {
        spellInfo = paladinSpellInfo.judgment,
        -- _debugName = "judgment_before_TV",
        tests = {
            { aType = HoJ.TYPE_POWER, aConst = 1, bType = HoJ.TYPE_CONSTANT, bConst = 4, op = HoJ.OP_EQUAL }, -- op = HoJ.OP_GREATER_EQUAL },
            { aType = HoJ.TYPE_TARGET_RANGE, aConst = paladinSpellInfo.judgment.range, bType = HoJ.TYPE_CONSTANT, bConst = true, op = HoJ.OP_EQUAL },
            { aType = HoJ.TYPE_RESULT, aConst = 1, bType = HoJ.TYPE_RESULT, bConst = 2, op = HoJ.OP_AND },
            { aType = HoJ.TYPE_TARGET_PLAYER_DEBUFF_REMAINING, aConst = paladinSpellInfo.judgmentDebuff.id, bType = HoJ.TYPE_CONSTANT, bConst = 0.0, op = HoJ.OP_LESS_EQUAL },
            { aType = HoJ.TYPE_RESULT, aConst = 3, bType = HoJ.TYPE_RESULT, bConst = 4, op = HoJ.OP_AND }
        }
    },
    {
        -- use TV if at 5 HoPo
        spellInfo = paladinSpellInfo.templarsVerdict,
        -- _debugName = "TV",
        tests = {
            { aType = HoJ.TYPE_POWER, aConst = 0, bType = HoJ.TYPE_CONSTANT, bConst = 5, op = HoJ.OP_GREATER_EQUAL },
            { aType = HoJ.TYPE_TARGET_RANGE, aConst = paladinSpellInfo.templarsVerdict.range, bType = HoJ.TYPE_CONSTANT, bConst = true, op = HoJ.OP_EQUAL },
            { aType = HoJ.TYPE_RESULT, aConst = 1, bType = HoJ.TYPE_RESULT, bConst = 2, op = HoJ.OP_AND },
        }
    },
    -- also use TV if we have Judgment on the target that is about to expire
    {
        spellInfo = paladinSpellInfo.templarsVerdict,
        -- _debugName = "TV_judgment_bonus_expiring",
        tests = {
            { aType = HoJ.TYPE_POWER, aConst = 0, bType = HoJ.TYPE_CONSTANT, bConst = 3, op = HoJ.OP_GREATER_EQUAL },
            { aType = HoJ.TYPE_TARGET_RANGE, aConst = paladinSpellInfo.templarsVerdict.range, bType = HoJ.TYPE_CONSTANT, bConst = true, op = HoJ.OP_EQUAL },
            { aType = HoJ.TYPE_RESULT, aConst = 1, bType = HoJ.TYPE_RESULT, bConst = 2, op = HoJ.OP_AND },
            { aType = HoJ.TYPE_TARGET_PLAYER_DEBUFF_REMAINING, aConst = paladinSpellInfo.judgmentDebuff.id, bType = HoJ.TYPE_CONSTANT, bConst = { 0.0, 2.0 }, op = HoJ.OP_IN_RANGE_MIN_EX_MAX_IN },
            { aType = HoJ.TYPE_RESULT, aConst = 3, bType = HoJ.TYPE_RESULT, bConst = 4, op = HoJ.OP_AND }
        }
    },
--[[
    -- wait on judgement if it is coming off of cooldown and we have at least 2 HoPo
    {
        spellInfo = paladinSpellInfo.judgment,
        _skipCooldownCheck = true,
        -- _debugName = "judgment_before_TV",
        tests = {
            { aType = HoJ.TYPE_POWER, aConst = 1, bType = HoJ.TYPE_CONSTANT, bConst = 2, op = HoJ.OP_GREATER_EQUAL },
            { aType = HoJ.TYPE_TARGET_RANGE, aConst = paladinSpellInfo.judgment.range, bType = HoJ.TYPE_CONSTANT, bConst = true, op = HoJ.OP_EQUAL },
            { aType = HoJ.TYPE_RESULT, aConst = 1, bType = HoJ.TYPE_RESULT, bConst = 2, op = HoJ.OP_AND },
            { aType = HoJ.TYPE_SKILL_COOLDOWN_REMAINING, aConst = paladinSpellInfo.judgment, bType = HoJ.TYPE_CONSTANT, bConst = 2, op = HoJ.OP_LESS_EQUAL },
            { aType = HoJ.TYPE_RESULT, aConst = 3, bType = HoJ.TYPE_RESULT, bConst = 4, op = HoJ.OP_AND },
        }
    },
--]]
    {
        -- this relies on the spell's minLevel check because getting spell rank since 8.0 is a PITA 
        -- and unreliabel and rank 2 of hammer of wrath is a whole new spell, anyway
        spellInfo = paladinSpellInfo.hammerOfWrath2,
        -- _debugName = "hammerOfWrath_Rank2",
        tests = {
            { aType = HoJ.TYPE_POWER, aConst = 0, bType = HoJ.TYPE_CONSTANT, bConst = 4, op = HoJ.OP_LESS_EQUAL },
            { aType = HoJ.TYPE_TARGET_RANGE, aConst = paladinSpellInfo.hammerOfWrath2.range, bType = HoJ.TYPE_CONSTANT, bConst = true, op = HoJ.OP_EQUAL },
            { aType = HoJ.TYPE_RESULT, aConst = 1, bType = HoJ.TYPE_RESULT, bConst = 2, op = HoJ.OP_AND },
            { aType = HoJ.TYPE_PLAYER_BUFF_ACTIVE, aConst = paladinSpellInfo.crusade.id, bType = HoJ.TYPE_CONSTANT, bConst = true, op = HoJ.OP_EQUAL },
            { aType = HoJ.TYPE_PLAYER_BUFF_ACTIVE, aConst = paladinSpellInfo.avengingWrath.id, bType = HoJ.TYPE_CONSTANT, bConst = true, op = HoJ.OP_EQUAL },
            { aType = HoJ.TYPE_RESULT, aConst = 4, bType = HoJ.TYPE_RESULT, bConst = 5, op = HoJ.OP_OR },
            { aType = HoJ.TYPE_RESULT, aConst = 3, bType = HoJ.TYPE_RESULT, bConst = 6, op = HoJ.OP_AND },
        }
    },
    {
        spellInfo = paladinSpellInfo.bladeOfJustice,
        -- _debugName = "bladeOfJustice",
        tests = {
            { aType = HoJ.TYPE_POWER, aConst = 0, bType = HoJ.TYPE_CONSTANT, bConst = 3, op = HoJ.OP_LESS_EQUAL },
            { aType = HoJ.TYPE_TARGET_RANGE, aConst = paladinSpellInfo.bladeOfJustice.range, bType = HoJ.TYPE_CONSTANT, bConst = true, op = HoJ.OP_EQUAL },
            { aType = HoJ.TYPE_RESULT, aConst = 1, bType = HoJ.TYPE_RESULT, bConst = 2, op = HoJ.OP_AND },
        }
    },
    {
        spellInfo = paladinSpellInfo.judgment,
        -- _debugName = "judgment",
        tests = {
            { aType = HoJ.TYPE_POWER, aConst = 0, bType = HoJ.TYPE_CONSTANT, bConst = 4, op = HoJ.OP_LESS_EQUAL },
            { aType = HoJ.TYPE_TARGET_RANGE, aConst = paladinSpellInfo.judgment.range, bType = HoJ.TYPE_CONSTANT, bConst = true, op = HoJ.OP_EQUAL },
            { aType = HoJ.TYPE_RESULT, aConst = 1, bType = HoJ.TYPE_RESULT, bConst = 2, op = HoJ.OP_AND },
        }
    },
    {
        -- use TV if at 3 HoPo if we less than 2 crusader strike charges
        spellInfo = paladinSpellInfo.templarsVerdict,
        -- _debugName = "TV",
        tests = {
            { aType = HoJ.TYPE_POWER, aConst = 0, bType = HoJ.TYPE_CONSTANT, bConst = 3, op = HoJ.OP_GREATER_EQUAL },
            { aType = HoJ.TYPE_SKILL_CHARGES, aConst = paladinSpellInfo.crusaderStrike, bType = HoJ.TYPE_CONSTANT, bConst = 2, op = HoJ.OP_LESS },
            { aType = HoJ.TYPE_RESULT, aConst = 1, bType = HoJ.TYPE_RESULT, bConst = 2, op = HoJ.OP_AND },
            { aType = HoJ.TYPE_TARGET_RANGE, aConst = paladinSpellInfo.templarsVerdict.range, bType = HoJ.TYPE_CONSTANT, bConst = true, op = HoJ.OP_EQUAL },
            { aType = HoJ.TYPE_RESULT, aConst = 3, bType = HoJ.TYPE_RESULT, bConst = 4, op = HoJ.OP_AND },
        }
    },
    {
        spellInfo = paladinSpellInfo.crusaderStrike,
        -- _debugName = "crusaderStrike",
        tests = {
            { aType = HoJ.TYPE_POWER, aConst = 0, bType = HoJ.TYPE_CONSTANT, bConst = 4, op = HoJ.OP_LESS_EQUAL },
            { aType = HoJ.TYPE_TARGET_RANGE, aConst = paladinSpellInfo.crusaderStrike.range, bType = HoJ.TYPE_CONSTANT, bConst = true, op = HoJ.OP_EQUAL },
            { aType = HoJ.TYPE_RESULT, aConst = 1, bType = HoJ.TYPE_RESULT, bConst = 2, op = HoJ.OP_AND }
        }
    },
    -- if no HoPo generators are available and we have >= 3 Holy Power or Divine Purpose, use TV / DS
    {
        spellInfo = paladinSpellInfo.templarsVerdict,
        -- _debugName = "templarsVerdict_DivinePurpose",
        tests = {
            { aType = HoJ.TYPE_POWER, aConst = 0, bType = HoJ.TYPE_CONSTANT, bConst = 3, op = HoJ.OP_GREATER_EQUAL },
            { aType = HoJ.TYPE_PLAYER_BUFF_ACTIVE, aConst = paladinSpellInfo.divinePurpose.id, bType = HoJ.TYPE_CONSTANT, bConst = true, op = HoJ.OP_EQUAL },
            { aType = HoJ.TYPE_RESULT, aConst = 1, bType = HoJ.TYPE_RESULT, bConst = 2, op = HoJ.OP_OR },
            { aType = HoJ.TYPE_TARGET_RANGE, aConst = paladinSpellInfo.templarsVerdict.range, bType = HoJ.TYPE_CONSTANT, bConst = true, op = HoJ.OP_EQUAL },
            { aType = HoJ.TYPE_RESULT, aConst = 3, bType = HoJ.TYPE_RESULT, bConst = 4, op = HoJ.OP_AND }
        }
    },
    {
        -- empyrean power if we have nothing else to use
        spellInfo = paladinSpellInfo.empyreanPower,
        -- _debugName = "empyreanPower_filler",
        tests = {
            { aType = HoJ.TYPE_PLAYER_BUFF_ACTIVE, aConst = paladinSpellInfo.empyreanPowerBuff.id, bType = HoJ.TYPE_CONSTANT, bConst = true, op = HoJ.OP_EQUAL },
            { aType = HoJ.TYPE_TARGET_RANGE, aConst = paladinSpellInfo.empyreanPower.range, bType = HoJ.TYPE_CONSTANT, bConst = true, op = HoJ.OP_EQUAL },
            { aType = HoJ.TYPE_RESULT, aConst = 1, bType = HoJ.TYPE_RESULT, bConst = 2, op = HoJ.OP_AND },
        }
    },
}

local retInterruptRules = {
    {
        spellInfo = paladinSpellInfo.rebuke,
        -- _debugName = "rebuke",
        tests = {
            { aType = HoJ.TYPE_TARGET_RANGE, aConst = paladinSpellInfo.rebuke.range, bType = HoJ.TYPE_CONSTANT, bConst = true, op = HoJ.OP_EQUAL },
            { aType = HoJ.TYPE_TARGET_CAN_BE_INTERRUPTED, aConst = 0, bType = HoJ.TYPE_CONSTANT, bConst = true, op = HoJ.OP_EQUAL },
            { aType = HoJ.TYPE_RESULT, aConst = 1, bType = HoJ.TYPE_RESULT, bConst = 2, op = HoJ.OP_AND },
        }
    },
    {
        spellInfo = paladinSpellInfo.blindingLight,
        -- _debugName = "blindingLight",
        tests = {
            { aType = HoJ.TYPE_TARGET_RANGE, aConst = paladinSpellInfo.blindingLight.range, bType = HoJ.TYPE_CONSTANT, bConst = true, op = HoJ.OP_EQUAL },
            { aType = HoJ.TYPE_TARGET_CAN_BE_INTERRUPTED, aConst = 0, bType = HoJ.TYPE_CONSTANT, bConst = true, op = HoJ.OP_EQUAL },
            { aType = HoJ.TYPE_RESULT, aConst = 1, bType = HoJ.TYPE_RESULT, bConst = 2, op = HoJ.OP_AND },
            { aType = HoJ.TYPE_TARGET_IS_BOSS, aConst = 0, bType = HoJ.TYPE_CONSTANT, bConst = false, op = HoJ.OP_EQUAL },
            { aType = HoJ.TYPE_RESULT, aConst = 3, bType = HoJ.TYPE_RESULT, bConst = 4, op = HoJ.OP_AND },
        }
    },
    {
        spellInfo = paladinSpellInfo.fistOfJustice,
        --_debugName = "fistOfJustice",
        tests = {
            { aType = HoJ.TYPE_TARGET_RANGE, aConst = paladinSpellInfo.hammerOfJustice.range, bType = HoJ.TYPE_CONSTANT, bConst = true, op = HoJ.OP_EQUAL },
            { aType = HoJ.TYPE_TARGET_IS_BOSS, aConst = 0, bType = HoJ.TYPE_CONSTANT, bConst = false, op = HoJ.OP_EQUAL },
            { aType = HoJ.TYPE_RESULT, aConst = 1, bType = HoJ.TYPE_RESULT, bConst = 2, op = HoJ.OP_AND },
            { aType = HoJ.TYPE_TARGET_IS_BOSS, aConst = 0, bType = HoJ.TYPE_CONSTANT, bConst = false, op = HoJ.OP_EQUAL },
            { aType = HoJ.TYPE_SKILL_OFF_COOLDOWN, aConst = paladinSpellInfo.hammerOfJustice, bType = HoJ.TYPE_CONSTANT, bConst = true, op = HoJ.OP_EQUAL },
            { aType = HoJ.TYPE_RESULT, aConst = 4, bType = HoJ.TYPE_RESULT, bConst = 5, op = HoJ.OP_AND },
            { aType = HoJ.TYPE_RESULT, aConst = 3, bType = HoJ.TYPE_RESULT, bConst = 6, op = HoJ.OP_AND },
        }
    },
    {
        spellInfo = paladinSpellInfo.hammerOfJustice,
        replacedByTalent = paladinSpellInfo.fistOfJustice,
        --_debugName = "hammerOfJustice",
        tests = {
            { aType = HoJ.TYPE_TARGET_RANGE, aConst = paladinSpellInfo.hammerOfJustice.range, bType = HoJ.TYPE_CONSTANT, bConst = true, op = HoJ.OP_EQUAL },
            { aType = HoJ.TYPE_TARGET_CAN_BE_INTERRUPTED, aConst = 0, bType = HoJ.TYPE_CONSTANT, bConst = true, op = HoJ.OP_EQUAL },
            { aType = HoJ.TYPE_RESULT, aConst = 1, bType = HoJ.TYPE_RESULT, bConst = 2, op = HoJ.OP_AND },
            { aType = HoJ.TYPE_TARGET_IS_BOSS, aConst = 0, bType = HoJ.TYPE_CONSTANT, bConst = false, op = HoJ.OP_EQUAL },
            { aType = HoJ.TYPE_RESULT, aConst = 3, bType = HoJ.TYPE_RESULT, bConst = 4, op = HoJ.OP_AND },
        }
    },
    {
        spellInfo = paladinSpellInfo.repentance,
        -- _debugName = "repentance",
        tests = {
            { aType = HoJ.TYPE_TARGET_RANGE, aConst = paladinSpellInfo.repentance.range, bType = HoJ.TYPE_CONSTANT, bConst = true, op = HoJ.OP_EQUAL },
            { aType = HoJ.TYPE_SPELL_CREATURE_TYPES, aConst = paladinSpellInfo.repentance, bType = HoJ.TYPE_TARGET_CREATURE_TYPE, bConst = 1, op = HoJ.OP_TABLE_CONTAINS },
            { aType = HoJ.TYPE_RESULT, aConst = 1, bType = HoJ.TYPE_RESULT, bConst = 2, op = HoJ.OP_AND },
        }
    },
}

local retMitigationRules = {
    -- emergency mitigations
    -- these are selected if:
    --   - the amount of damage predicted over the next 5 seconds exceeds the player's current health
    --   - predicted death time is < 5 seconds
    -- NOTE: non-instant healing is not considered a mitigation
    {
        spellInfo = paladinSpellInfo.layOnHands,
        -- _debugSuccessFn = HoJ.DebugRule_LayOnHands,
        -- _debugName = "layOnHands_emergency",
        tests = {
            { aType = HoJ.TYPE_PLAYER_TIME_TO_LIVE, aConst = 0, bType = HoJ.TYPE_CONSTANT, bConst = 10.0, op = HoJ.OP_LESS_EQUAL },
            { aType = HoJ.TYPE_NET_INCOMING_DPS_CHR, aConst = 5.0, bType = HoJ.TYPE_PLAYER_HEALTH_RATIO, bConst = 0 , op = HoJ.OP_GREATER_EQUAL },
            { aType = HoJ.TYPE_RESULT, aConst = 1, bType = HoJ.TYPE_RESULT, bConst = 2, op = HoJ.OP_OR },
            { aType = HoJ.TYPE_PLAYER_BUFF_ACTIVE, aConst = paladinSpellInfo.forbearance.id, bType = HoJ.TYPE_CONSTANT, bConst = false, op = HoJ.OP_EQUAL },
            { aType = HoJ.TYPE_RESULT, aConst = 3, bType = HoJ.TYPE_RESULT, bConst = 4, op = HoJ.OP_AND },
        }
    },
    {
        spellInfo = paladinSpellInfo.blessingOfProtection,
        -- _debugName = "blessingOfProtection",
        tests = {
            { aType = HoJ.TYPE_PLAYER_TIME_TO_LIVE, aConst = 0, bType = HoJ.TYPE_CONSTANT, bConst = 5.0, op = HoJ.OP_LESS },
            { aType = HoJ.TYPE_NET_INCOMING_DPS_CHR, aConst = 5.0, bType = HoJ.TYPE_PLAYER_HEALTH_RATIO, bConst = 0 , op = HoJ.OP_GREATER_EQUAL },
            { aType = HoJ.TYPE_RESULT, aConst = 1, bType = HoJ.TYPE_RESULT, bConst = 2, op = HoJ.OP_OR },
            { aType = HoJ.TYPE_SECONDS_SINCE_DAMAGED_BY_SPELL, aConst = 0, bType = HoJ.TYPE_CONSTANT, bConst = 10.0, op = HoJ.OP_LESS },
            { aType = HoJ.TYPE_PLAYER_BUFF_ACTIVE, aConst = paladinSpellInfo.forbearance.id, bType = HoJ.TYPE_CONSTANT, bConst = false, op = HoJ.OP_EQUAL },
            { aType = HoJ.TYPE_RESULT, aConst = 4, bType = HoJ.TYPE_RESULT, bConst = 5, op = HoJ.OP_AND },
            { aType = HoJ.TYPE_RESULT, aConst = 3, bType = HoJ.TYPE_RESULT, bConst = 6, op = HoJ.OP_AND },
        }
    },
    {
        spellInfo = paladinSpellInfo.divineShield,
        -- _debugName = "divineShield_emergency",
        tests = {
            { aType = HoJ.TYPE_PLAYER_TIME_TO_LIVE, aConst = 0, bType = HoJ.TYPE_CONSTANT, bConst = 5.0, op = HoJ.OP_LESS },
            { aType = HoJ.TYPE_NET_INCOMING_DPS_CHR, aConst = 5.0, bType = HoJ.TYPE_PLAYER_HEALTH_RATIO, bConst = 0 , op = HoJ.OP_GREATER_EQUAL },
            { aType = HoJ.TYPE_RESULT, aConst = 1, bType = HoJ.TYPE_RESULT, bConst = 2, op = HoJ.OP_OR },
            { aType = HoJ.TYPE_PLAYER_BUFF_ACTIVE, aConst = paladinSpellInfo.forbearance.id, bType = HoJ.TYPE_CONSTANT, bConst = false, op = HoJ.OP_EQUAL },
            { aType = HoJ.TYPE_RESULT, aConst = 3, bType = HoJ.TYPE_RESULT, bConst = 4, op = HoJ.OP_AND },
        }
    },
    {
        spellInfo = paladinSpellInfo.shieldOfVengeance,
        -- _debugName = "shieldOfVengeance_emergency",
        tests = {
            { aType = HoJ.TYPE_PLAYER_TIME_TO_LIVE, aConst = 0, bType = HoJ.TYPE_CONSTANT, bConst = 5.0, op = HoJ.OP_LESS },
            { aType = HoJ.TYPE_NET_INCOMING_DPS_CHR, aConst = 5.0, bType = HoJ.TYPE_PLAYER_HEALTH_RATIO, bConst = 0 , op = HoJ.OP_GREATER_EQUAL },
            { aType = HoJ.TYPE_RESULT, aConst = 1, bType = HoJ.TYPE_RESULT, bConst = 2, op = HoJ.OP_OR },
        }
    },
    -- moderate mitigations
    -- these are selected if:
    --   - the damage per second ratio (ratio of damage to player max health) is greater than 0.1 (i.e. 10% of health per second)
    --   - the damage per second over the next 5 seconds exceeds the player's current health ratio
    {
        spellInfo = paladinSpellInfo.shieldOfVengeance,
        --_debugName = "shieldOfVengeance",
        tests = {
            { aType = HoJ.TYPE_PLAYER_TIME_TO_LIVE, aConst = 0, bType = HoJ.TYPE_CONSTANT, bConst = 30.0, op = HoJ.OP_LESS_EQUAL },
            { aType = HoJ.TYPE_NET_INCOMING_DPS_CHR, aConst = 1.0, bType = HoJ.TYPE_CONSTANT, bConst = 0.1, op = HoJ.OP_GREATER_EQUAL },
            { aType = HoJ.TYPE_RESULT, aConst = 1, bType = HoJ.TYPE_RESULT, bConst = 2, op = HoJ.OP_OR },
        }
    },
    {
        spellInfo = paladinSpellInfo.blessingOfFreedom,
        -- _debugName = "blessingOfFreedom",
        tests = {
            { aType = HoJ.TYPE_UNIT_IS_SLOWED, aConst = "player", bType = HoJ.TYPE_CONSTANT, bConst = true, op = HoJ.OP_EQUAL },
        }
    },
}

local retDpsCooldownRules = {
    {
        spellInfo = paladinSpellInfo.crusade,
        --_debugName = "crusade",
        tests = {
            { aType = HoJ.TYPE_COMBAT_TARGET_TIME_TO_LIVE, aConst = 1, bType = HoJ.TYPE_SPELL_DURATION, bConst = paladinSpellInfo.crusade, op = HoJ.OP_GREATER },
            { aType = HoJ.TYPE_COMBAT_ATTACKS_HIT, aConst = 0, bType = HoJ.TYPE_CONSTANT, bConst = 1, op = HoJ.OP_GREATER },
            { aType = HoJ.TYPE_RESULT, aConst = 1, bType = HoJ.TYPE_RESULT, bConst = 2, op = HoJ.OP_AND },
            { aType = HoJ.TYPE_TARGET_IS_TOUGH, aConst = 0, bType = HoJ.TYPE_CONSTANT, bConst = true, op = HoJ.OP_EQUAL },
            { aType = HoJ.TYPE_RESULT, aConst = 3, bType = HoJ.TYPE_RESULT, bConst = 4, op = HoJ.OP_OR },
            { aType = HoJ.TYPE_PLAYER_BUFF_ACTIVE, aConst = paladinSpellInfo.crusade.id, bType = HoJ.TYPE_CONSTANT, bConst = false, op = HoJ.OP_EQUAL },
            { aType = HoJ.TYPE_RESULT, aConst = 5, bType = HoJ.TYPE_RESULT, bConst = 6, op = HoJ.OP_AND },
        }
    },
    {
        spellInfo = paladinSpellInfo.avengingWrath,
        replacedByTalent = paladinSpellInfo.crusade,
        -- _debugName = "avengingWrath",
        tests = {
            { aType = HoJ.TYPE_COMBAT_TARGET_TIME_TO_LIVE, aConst = 1, bType = HoJ.TYPE_SPELL_DURATION, bConst = paladinSpellInfo.avengingWrath, op = HoJ.OP_GREATER },
            { aType = HoJ.TYPE_COMBAT_ATTACKS_HIT, aConst = 0, bType = HoJ.TYPE_CONSTANT, bConst = 1, op = HoJ.OP_GREATER },
            { aType = HoJ.TYPE_RESULT, aConst = 1, bType = HoJ.TYPE_RESULT, bConst = 2, op = HoJ.OP_AND },
            { aType = HoJ.TYPE_TARGET_IS_TOUGH, aConst = 0, bType = HoJ.TYPE_CONSTANT, bConst = true, op = HoJ.OP_EQUAL },
            { aType = HoJ.TYPE_RESULT, aConst = 3, bType = HoJ.TYPE_RESULT, bConst = 4, op = HoJ.OP_OR },
            { aType = HoJ.TYPE_PLAYER_BUFF_ACTIVE, aConst = paladinSpellInfo.avengingWrath.id, bType = HoJ.TYPE_CONSTANT, bConst = false, op = HoJ.OP_EQUAL },
            { aType = HoJ.TYPE_RESULT, aConst = 5, bType = HoJ.TYPE_RESULT, bConst = 6, op = HoJ.OP_AND },

        }
    },
    {
        spellInfo = paladinSpellInfo.wordOfGlory,
        -- _debugName = "wordOfGlory",
        tests = {
            { aType = HoJ.TYPE_PLAYER_TIME_TO_LIVE, aConst = 1, bType = HoJ.TYPE_CONSTANT, bConst = 5.0, op = HoJ.OP_LESS },
            { aType = HoJ.TYPE_PLAYER_HEALTH_RATIO, aConst = 0, bType = HoJ.TYPE_CONSTANT, bConst = 0.6, op = HoJ.OP_LESS },
            { aType = HoJ.TYPE_RESULT, aConst = 1, bType = HoJ.TYPE_RESULT, bConst = 2, op = HoJ.OP_OR },
            { aType = HoJ.TYPE_PLAYER_BUFF_ACTIVE, aConst = paladinSpellInfo.divinePurposeBuff.id, bType = HoJ.TYPE_CONSTANT, bConst = true, op = HoJ.OP_EQUAL },
            { aType = HoJ.TYPE_POWER, aConst = 1, bType = HoJ.TYPE_CONSTANT, bConst = 3, op = HoJ.OP_GREATER_EQUAL },
            { aType = HoJ.TYPE_RESULT, aConst = 4, bType = HoJ.TYPE_RESULT, bConst = 5, op = HoJ.OP_OR },
            { aType = HoJ.TYPE_PLAYER_BUFF_STACK_COUNT, aConst = paladinSpellInfo.selflessHealerBuff, bType = HoJ.TYPE_CONSTANT, bConst = 4, op = HoJ.OP_LESS },
            { aType = HoJ.TYPE_RESULT, aConst = 3, bType = HoJ.TYPE_RESULT, bConst = 6, op = HoJ.OP_AND },
            { aType = HoJ.TYPE_RESULT, aConst = 7, bType = HoJ.TYPE_RESULT, bConst = 8, op = HoJ.OP_AND },
        }
    },
}

-- every possible spell that can show on the cooldown bar must have a priorty in this list
local retCooldownSortPriorities = {
    _name = "retCooldownSortPriorities",
    [paladinSpellInfo.layOnHands.id] = 13,
    [paladinSpellInfo.divineShield.id] = 12,
    [paladinSpellInfo.blessingOfProtection.id] = 11,
    [paladinSpellInfo.wordOfGlory.id] = 10,
    [paladinSpellInfo.rebuke.id] = 9,
    [paladinSpellInfo.repentance.id] = 8,
    [paladinSpellInfo.hammerOfJustice.id] = 7,
    [paladinSpellInfo.fistOfJustice.id] = 6,
    [paladinSpellInfo.blindingLight.id] = 5,
    [paladinSpellInfo.shieldOfVengeance.id] = 4,
    [paladinSpellInfo.crusade.id] = 3,
    [paladinSpellInfo.avengingWrath.id] = 2,
    [paladinSpellInfo.blessingOfFreedom.id] = 1,
}

local retActiveBuffList = {
    [paladinSpellInfo.giftOfTheNaaru.id] = { isActive = false, start = 0, duration = 0, timeRemaining = 0, stacks = 0 },
    [paladinSpellInfo.layOnHands.id] = { isActive = false, start = 0, duration = 0, timeRemaining = 0, stacks = 0 },
    [paladinSpellInfo.blessingOfProtection.id] = { isActive = false, start = 0, duration = 0, timeRemaining = 0, stacks = 0 },
    [paladinSpellInfo.divineShield.id] = { isActive = false, start = 0, duration = 0, timeRemaining = 0, stacks = 0 },
    [paladinSpellInfo.forbearance.id] = { isActive = false, start = 0, duration = 0, timeRemaining = 0, stacks = 0 },
    [paladinSpellInfo.shieldOfVengeance.id] = { isActive = false, start = 0, duration = 0, timeRemaining = 0, stacks = 0 },
    [paladinSpellInfo.avengingWrath.id] = { isActive = false, start = 0, duration = 0, timeRemaining = 0, stacks = 0 },
    [paladinSpellInfo.crusade.id] = { isActive = false, start = 0, duration = 0, timeRemaining = 0, stacks = 0 },
    [paladinSpellInfo.consecration.id] = { isActive = false, start = 0, duration = 0, timeRemaining = 0, stacks = 0 },
    [paladinSpellInfo.divinePurposeBuff.id] = { isActive = false, start = 0, duration = 0, timeRemaining = 0, stacks = 0 },
    [paladinSpellInfo.inquisition.id] = { isActive = false, start = 0, duration = 0, timeRemaining = 0, stacks = 0 },
    [paladinSpellInfo.empyreanPowerBuff.id] = { isActive = false, start = 0, duration = 0, timeRemaining = 0, stacks = 0 },
	[paladinSpellInfo.selflessHealerBuff.id] = { isActive = false, start = 0, duration = 0, timeRemaining = 0, stacks = 0 },
	[paladinSpellInfo.righteousVerdictBuff.id] = { isActive = false, start = 0, duration = 0, timeRemaining = 0, stacks = 0 },
}

local retActiveBlessingsList = {
    -- SLREMOVE    
--[[
    [paladinSpellInfo.blessingOfKings.id] = {
        spellInfo = paladinSpellInfo.blessingOfKings,
        targetName = nil,
        start = 0.0,
        duration = 0.0,
        updateIndex = -1
    },
    [paladinSpellInfo.blessingOfWisdom.id] = {
        spellInfo = paladinSpellInfo.blessingOfWisdom,
        targetName = nil,
        start = 0.0,
        duration = 0.0,
        updateIndex = -1
    }
--]]
}

local retProcdBuffList = {
	{ _spellInfo = paladinSpellInfo.righteousVerdictBuff },
    { _spellInfo = paladinSpellInfo.divinePurposeBuff },
    { _spellInfo = paladinSpellInfo.empyreanPowerBuff },
    { _spellInfo = paladinSpellInfo.selflessHealerBuff },
    { _spellInfo = paladinSpellInfo.inquisition },
    { _spellInfo = paladinSpellInfo.consecration },
}

local retBuffSortPriorities = {
	_name = "retBuffSortPriorities",
	[paladinSpellInfo.righteousVerdictBuff.id] = 7,
    [paladinSpellInfo.selflessHealerBuff.id] = 6,
    [paladinSpellInfo.divinePurposeBuff.id] = 5,
    [paladinSpellInfo.empyreanPowerBuff.id] = 4,
    [paladinSpellInfo.inquisition.id] = 3,
    [paladinSpellInfo.consecration.id] = 2,
    -- SLREMOVE
--[[
    [paladinSpellInfo.blessingOfKings.id] = 2,
    [paladinSpellInfo.blessingOfWisdom.id] = 1,
--]]
}

local retAoeSpellTable = {
    [paladinSpellInfo.consecration.id] = {
        _buffId = paladinSpellInfo.consecration.id
    }
}

local retAoeBuffTable = {
    [paladinSpellInfo.consecration.id] = {
        _castById = paladinSpellInfo.consecration,
        _spellInfo = paladinSpellInfo.consecration,
        _startTime = 0.0,
        _duration = 0.0,
        _expiration = 0.0,
    }
}


--------------------------------------------------------------------------------
-- PROTECTION
--------------------------------------------------------------------------------

local protRotationRules = {
    _name = "protRotationRules",
    {
        spellInfo = paladinSpellInfo.seraphim,
        _debugName = "seraphim",
        -- _debugSuccess = true,
        -- _debugFail = true,
        tests = {
            { aType = HoJ.TYPE_PLAYER_BUFF_ACTIVE, aConst = paladinSpellInfo.divinePurposeBuff.id, bType = HoJ.TYPE_CONSTANT, bConst = true, op = HoJ.OP_EQUAL },
            { aType = HoJ.TYPE_POWER, aConst = 1, bType = HoJ.TYPE_CONSTANT, bConst = 3, op = HoJ.OP_GREATER_EQUAL },
            { aType = HoJ.TYPE_RESULT, aConst = 1, bType = HoJ.TYPE_RESULT, bConst = 2, op = HoJ.OP_OR },
            
                { aType = HoJ.TYPE_TARGET_IS_BOSS, aConst = 0, bType = HoJ.TYPE_CONSTANT, bConst = false, op = HoJ.OP_EQUAL },
                { aType = HoJ.TYPE_TARGET_IS_TOUGH, aConst = 0, bType = HoJ.TYPE_CONSTANT, bConst = true, op = HoJ.OP_EQUAL },
                { aType = HoJ.TYPE_RESULT, aConst = 4, bType = HoJ.TYPE_RESULT, bConst = 5, op = HoJ.OP_OR },

                { aType = HoJ.TYPE_COMBAT_TARGET_TIME_TO_LIVE, aConst = 1, bType = HoJ.TYPE_SPELL_DURATION, bConst = paladinSpellInfo.seraphim, op = HoJ.OP_GREATER },
                { aType = HoJ.TYPE_COMBAT_ATTACKS_HIT, aConst = 0, bType = HoJ.TYPE_CONSTANT, bConst = 5, op = HoJ.OP_GREATER },
                { aType = HoJ.TYPE_RESULT, aConst = 7, bType = HoJ.TYPE_RESULT, bConst = 8, op = HoJ.OP_AND },

            { aType = HoJ.TYPE_RESULT, aConst = 6, bType = HoJ.TYPE_RESULT, bConst = 9, op = HoJ.OP_OR },
            { aType = HoJ.TYPE_RESULT, aConst = 3, bType = HoJ.TYPE_RESULT, bConst = 10, op = HoJ.OP_AND },
        }
    },
    {
        spellInfo = paladinSpellInfo.consecration,
        -- _debugName = "consecration",
        tests = {
            { aType = HoJ.TYPE_PLAYER_BUFF_ACTIVE, aConst = paladinSpellInfo.protConsecrationBuff.id, bType = HoJ.TYPE_CONSTANT, bConst = false, op = HoJ.OP_EQUAL },
        }
    },
    {
        spellInfo = paladinSpellInfo.judgment,
        -- _debugName = "judgment",
        tests = {
            { aType = HoJ.TYPE_TARGET_RANGE, aConst = paladinSpellInfo.judgment.range, bType = HoJ.TYPE_CONSTANT, bConst = true, op = HoJ.OP_EQUAL },
        }
    },
    {
        spellInfo = paladinSpellInfo.avengersShield,
        -- _debugName = "avengersShield",
        tests = {
            { aType = HoJ.TYPE_TARGET_RANGE, aConst = paladinSpellInfo.avengersShield.range, bType = HoJ.TYPE_CONSTANT, bConst = true, op = HoJ.OP_EQUAL },
        }
    },
    {
        spellInfo = paladinSpellInfo.blessedHammer,
        -- _debugName = "avengersShield",
        tests = {
            { aType = HoJ.TYPE_TARGET_RANGE, aConst = paladinSpellInfo.blessedHammer.range, bType = HoJ.TYPE_CONSTANT, bConst = true, op = HoJ.OP_EQUAL },
        }
    },
    {
        spellInfo = paladinSpellInfo.hammerOfTheRighteous,
        -- _debugName = "avengersShield",
        tests = {
            { aType = HoJ.TYPE_TALENT_SELECTED, aConst = paladinSpellInfo.blessedHammer, bType = HoJ.TYPE_CONSTANT, bConst = false, op = HoJ.OP_EQUAL },
            { aType = HoJ.TYPE_TARGET_RANGE, aConst = paladinSpellInfo.hammerOfTheRighteous.range, bType = HoJ.TYPE_CONSTANT, bConst = true, op = HoJ.OP_EQUAL },
        }
    },
    {
        spellInfo = paladinSpellInfo.consecration,
        -- _debugName = "consecration_if_off_cd",
        tests = {
            { aType = HoJ.TYPE_CONSTANT, aConst = 1, bType = HoJ.TYPE_CONSTANT, bConst = 1, op = HoJ.OP_EQUAL },
        }
    },
}

local protInterruptRules = {
    _name = "protInterruptRules",
    {
        spellInfo = paladinSpellInfo.rebuke,
        -- _debugName = "rebuke",
        tests = {
            { aType = HoJ.TYPE_TARGET_RANGE, aConst = paladinSpellInfo.rebuke.range, bType = HoJ.TYPE_CONSTANT, bConst = true, op = HoJ.OP_EQUAL },
            { aType = HoJ.TYPE_TARGET_CAN_BE_INTERRUPTED, aConst = 0, bType = HoJ.TYPE_CONSTANT, bConst = true, op = HoJ.OP_EQUAL },
            { aType = HoJ.TYPE_RESULT, aConst = 1, bType = HoJ.TYPE_RESULT, bConst = 2, op = HoJ.OP_AND },
        }
    },
    {
        spellInfo = paladinSpellInfo.avengersShield,
        -- _debugName = "avengersShield",
        tests = {
            { aType = HoJ.TYPE_TARGET_RANGE, aConst = paladinSpellInfo.avengersShield.range, bType = HoJ.TYPE_CONSTANT, bConst = true, op = HoJ.OP_EQUAL },
            { aType = HoJ.TYPE_TARGET_CAN_BE_INTERRUPTED, aConst = 0, bType = HoJ.TYPE_CONSTANT, bConst = true, op = HoJ.OP_EQUAL },
            { aType = HoJ.TYPE_RESULT, aConst = 1, bType = HoJ.TYPE_RESULT, bConst = 2, op = HoJ.OP_AND },
        }
    },
    {
        spellInfo = paladinSpellInfo.blindingLight,
        -- _debugName = "blindingLight",
        tests = {
            { aType = HoJ.TYPE_TARGET_RANGE, aConst = paladinSpellInfo.blindingLight.range, bType = HoJ.TYPE_CONSTANT, bConst = true, op = HoJ.OP_EQUAL },
            { aType = HoJ.TYPE_TARGET_CAN_BE_INTERRUPTED, aConst = 0, bType = HoJ.TYPE_CONSTANT, bConst = true, op = HoJ.OP_EQUAL },
            { aType = HoJ.TYPE_RESULT, aConst = 1, bType = HoJ.TYPE_RESULT, bConst = 2, op = HoJ.OP_AND },
            { aType = HoJ.TYPE_TARGET_IS_BOSS, aConst = 0, bType = HoJ.TYPE_CONSTANT, bConst = false, op = HoJ.OP_EQUAL },
            { aType = HoJ.TYPE_RESULT, aConst = 3, bType = HoJ.TYPE_RESULT, bConst = 4, op = HoJ.OP_AND },
        }
    },
    {
        spellInfo = paladinSpellInfo.fistOfJustice,
        --_debugName = "fistOfJustice",
        tests = {
            { aType = HoJ.TYPE_TARGET_RANGE, aConst = paladinSpellInfo.hammerOfJustice.range, bType = HoJ.TYPE_CONSTANT, bConst = true, op = HoJ.OP_EQUAL },
            { aType = HoJ.TYPE_TARGET_IS_BOSS, aConst = 0, bType = HoJ.TYPE_CONSTANT, bConst = false, op = HoJ.OP_EQUAL },
            { aType = HoJ.TYPE_RESULT, aConst = 1, bType = HoJ.TYPE_RESULT, bConst = 2, op = HoJ.OP_AND },
            { aType = HoJ.TYPE_TARGET_IS_BOSS, aConst = 0, bType = HoJ.TYPE_CONSTANT, bConst = false, op = HoJ.OP_EQUAL },
            { aType = HoJ.TYPE_SKILL_OFF_COOLDOWN, aConst = paladinSpellInfo.hammerOfJustice, bType = HoJ.TYPE_CONSTANT, bConst = true, op = HoJ.OP_EQUAL },
            { aType = HoJ.TYPE_RESULT, aConst = 4, bType = HoJ.TYPE_RESULT, bConst = 5, op = HoJ.OP_AND },
            { aType = HoJ.TYPE_RESULT, aConst = 3, bType = HoJ.TYPE_RESULT, bConst = 6, op = HoJ.OP_AND },
        }
    },
    {
        spellInfo = paladinSpellInfo.hammerOfJustice,
        replacedByTalent = paladinSpellInfo.fistOfJustice,
        --_debugName = "hammerOfJustice",
        tests = {
            { aType = HoJ.TYPE_TARGET_RANGE, aConst = paladinSpellInfo.hammerOfJustice.range, bType = HoJ.TYPE_CONSTANT, bConst = true, op = HoJ.OP_EQUAL },
            { aType = HoJ.TYPE_TARGET_CAN_BE_INTERRUPTED, aConst = 0, bType = HoJ.TYPE_CONSTANT, bConst = true, op = HoJ.OP_EQUAL },
            { aType = HoJ.TYPE_RESULT, aConst = 1, bType = HoJ.TYPE_RESULT, bConst = 2, op = HoJ.OP_AND },
            { aType = HoJ.TYPE_TARGET_IS_BOSS, aConst = 0, bType = HoJ.TYPE_CONSTANT, bConst = false, op = HoJ.OP_EQUAL },
            { aType = HoJ.TYPE_RESULT, aConst = 3, bType = HoJ.TYPE_RESULT, bConst = 4, op = HoJ.OP_AND },
        }
    },
    {
        spellInfo = paladinSpellInfo.repentance,
        -- _debugName = "repentance",
        tests = {
            { aType = HoJ.TYPE_TARGET_RANGE, aConst = paladinSpellInfo.repentance.range, bType = HoJ.TYPE_CONSTANT, bConst = true, op = HoJ.OP_EQUAL },
            { aType = HoJ.TYPE_SPELL_CREATURE_TYPES, aConst = paladinSpellInfo.repentance, bType = HoJ.TYPE_TARGET_CREATURE_TYPE, bConst = 1, op = HoJ.OP_TABLE_CONTAINS },
            { aType = HoJ.TYPE_RESULT, aConst = 1, bType = HoJ.TYPE_RESULT, bConst = 2, op = HoJ.OP_AND },
        }
    },
}

local protMitigationRules = {
    _name = "protMitigationRules",
    -- emergency mitigations
    -- NOTE: non-instant healing is not considered a mitigation
    {
        spellInfo = paladinSpellInfo.wordOfGlory,
        -- _debugSuccess = true,
        tests = {
            { aType = HoJ.TYPE_NET_INCOMING_DPS_CHR, aConst = 1.0, bType = HoJ.TYPE_CONSTANT, bConst = 0.05 , op = HoJ.OP_GREATER_EQUAL },
            { aType = HoJ.TYPE_PLAYER_HEALTH_RATIO, aConst = 0, bType = HoJ.TYPE_CONSTANT, bConst = 0.5, op = HoJ.OP_LESS_EQUAL },
            { aType = HoJ.TYPE_RESULT, aConst = 1, bType = HoJ.TYPE_RESULT, bConst = 2, op = HoJ.OP_OR },
            { aType = HoJ.TYPE_POWER, aConst = 1, bType = HoJ.TYPE_CONSTANT, bConst = 3, op = HoJ.OP_GREATER_EQUAL },
            { aType = HoJ.TYPE_PLAYER_BUFF_ACTIVE, aConst = paladinSpellInfo.divinePurposeBuff.id, bType = HoJ.TYPE_CONSTANT, bConst = true, op = HoJ.OP_EQUAL },
            { aType = HoJ.TYPE_RESULT, aConst = 4, bType = HoJ.TYPE_RESULT, bConst = 5, op = HoJ.OP_OR },
            { aType = HoJ.TYPE_PLAYER_BUFF_STACK_COUNT, aConst = paladinSpellInfo.shiningLightBuff, bType = HoJ.TYPE_CONSTANT, bConst = 3, op = HoJ.OP_EQUAL },
            { aType = HoJ.TYPE_RESULT, aConst = 6, bType = HoJ.TYPE_RESULT, bConst = 7, op = HoJ.OP_OR },
            { aType = HoJ.TYPE_RESULT, aConst = 3, bType = HoJ.TYPE_RESULT, bConst = 8, op = HoJ.OP_AND },
        }
    },
    {
        spellInfo = paladinSpellInfo.layOnHands,
        -- _debugName = "layOnHands_emergency",
        tests = {
            { aType = HoJ.TYPE_NET_INCOMING_DPS_CHR, aConst = 1.0, bType = HoJ.TYPE_CONSTANT, bConst = 0.1 , op = HoJ.OP_GREATER_EQUAL },
            { aType = HoJ.TYPE_PLAYER_HEALTH_RATIO, aConst = 0, bType = HoJ.TYPE_CONSTANT, bConst = 0.4, op = HoJ.OP_LESS_EQUAL },
            { aType = HoJ.TYPE_RESULT, aConst = 1, bType = HoJ.TYPE_RESULT, bConst = 2, op = HoJ.OP_AND },
            { aType = HoJ.TYPE_PLAYER_BUFF_ACTIVE, aConst = paladinSpellInfo.forbearance.id, bType = HoJ.TYPE_CONSTANT, bConst = false, op = HoJ.OP_EQUAL },
            { aType = HoJ.TYPE_RESULT, aConst = 3, bType = HoJ.TYPE_RESULT, bConst = 4, op = HoJ.OP_AND },
        }
    },
    {
        spellInfo = paladinSpellInfo.ardentDefender,
        -- _debugName = "ardentDefender_emergency",
        tests = {
            { aType = HoJ.TYPE_NET_INCOMING_DPS_CHR, aConst = 1.0, bType = HoJ.TYPE_CONSTANT, bConst = 0.1 , op = HoJ.OP_GREATER_EQUAL },
            { aType = HoJ.TYPE_PLAYER_HEALTH_RATIO, aConst = 0, bType = HoJ.TYPE_CONSTANT, bConst = 0.33, op = HoJ.OP_LESS_EQUAL },
            { aType = HoJ.TYPE_RESULT, aConst = 1, bType = HoJ.TYPE_RESULT, bConst = 2, op = HoJ.OP_AND },
        }
    },
    {
        spellInfo = paladinSpellInfo.guardianOfAncientKings,
        -- _debugName = "guardianOfAncientKings_emergency",
        tests = {
            { aType = HoJ.TYPE_NET_INCOMING_DPS_CHR, aConst = 1.0, bType = HoJ.TYPE_CONSTANT, bConst = 0.1 , op = HoJ.OP_GREATER_EQUAL },
            { aType = HoJ.TYPE_PLAYER_HEALTH_RATIO, aConst = 0, bType = HoJ.TYPE_CONSTANT, bConst = 0.2, op = HoJ.OP_LESS_EQUAL },
            { aType = HoJ.TYPE_RESULT, aConst = 1, bType = HoJ.TYPE_RESULT, bConst = 2, op = HoJ.OP_AND },
        }
    },
    {
        spellInfo = paladinSpellInfo.blessingOfProtection,
        -- _debugName = "blessingOfProtection",
        tests = {
            { aType = HoJ.TYPE_PLAYER_TIME_TO_LIVE, aConst = 0, bType = HoJ.TYPE_CONSTANT, bConst = 5.0, op = HoJ.OP_LESS },
            { aType = HoJ.TYPE_SECONDS_SINCE_DAMAGED_BY_SPELL, aConst = 0, bType = HoJ.TYPE_CONSTANT, bConst = 10.0, op = HoJ.OP_LESS },
            { aType = HoJ.TYPE_RESULT, aConst = 1, bType = HoJ.TYPE_RESULT, bConst = 2, op = HoJ.OP_AND },
            { aType = HoJ.TYPE_PLAYER_BUFF_ACTIVE, aConst = paladinSpellInfo.forbearance.id, bType = HoJ.TYPE_CONSTANT, bConst = false, op = HoJ.OP_EQUAL },
            { aType = HoJ.TYPE_RESULT, aConst = 3, bType = HoJ.TYPE_RESULT, bConst = 4, op = HoJ.OP_AND },
        }
    },
    {
        spellInfo = paladinSpellInfo.divineShield,
        -- _debugName = "divineShield",
        tests = {
            { aType = HoJ.TYPE_PLAYER_TIME_TO_LIVE, aConst = 0, bType = HoJ.TYPE_CONSTANT, bConst = 5.0, op = HoJ.OP_LESS },
            { aType = HoJ.TYPE_PLAYER_BUFF_ACTIVE, aConst = paladinSpellInfo.forbearance.id, bType = HoJ.TYPE_CONSTANT, bConst = false, op = HoJ.OP_EQUAL },
            { aType = HoJ.TYPE_RESULT, aConst = 1, bType = HoJ.TYPE_RESULT, bConst = 2, op = HoJ.OP_AND },
        }
    },
    -- moderate mitigations
    {
        spellInfo = paladinSpellInfo.ardentDefender,
        -- _debugName = "ardentDefender",
        tests = {
            { aType = HoJ.TYPE_NET_INCOMING_DPS_CHR, aConst = 10.0, bType = HoJ.TYPE_CONSTANT, bConst = 0.9 , op = HoJ.OP_GREATER_EQUAL },
            { aType = HoJ.TYPE_PLAYER_TIME_TO_LIVE, aConst = 0, bType = HoJ.TYPE_CONSTANT, bConst = 10.0, op = HoJ.OP_LESS },
            { aType = HoJ.TYPE_RESULT, aConst = 1, bType = HoJ.TYPE_RESULT, bConst = 2, op = HoJ.OP_OR },
        }
    },
    {
        spellInfo = paladinSpellInfo.guardianOfAncientKings,
        -- _debugName = "guardianOfAncientKings",
        tests = {
            { aType = HoJ.TYPE_NET_INCOMING_DPS_CHR, aConst = 10.0, bType = HoJ.TYPE_CONSTANT, bConst = 0.9 , op = HoJ.OP_GREATER_EQUAL },
            { aType = HoJ.TYPE_PLAYER_TIME_TO_LIVE, aConst = 0, bType = HoJ.TYPE_CONSTANT, bConst = 10.0, op = HoJ.OP_LESS },
            { aType = HoJ.TYPE_RESULT, aConst = 1, bType = HoJ.TYPE_RESULT, bConst = 2, op = HoJ.OP_OR },
        }
    },
    {
        spellInfo = paladinSpellInfo.shieldOfTheRighteous,
        -- _debugName = "shieldOfTheRighteous",
        tests = {
            { aType = HoJ.TYPE_POWER, aConst = 0, bType = HoJ.TYPE_CONSTANT, bConst = 3, op = HoJ.OP_GREATER_EQUAL },
            { aType = HoJ.TYPE_PLAYER_BUFF_ACTIVE, aConst = paladinSpellInfo.shieldOfTheRighteousBuff.id, bType = HoJ.TYPE_CONSTANT, bConst = false, op = HoJ.OP_EQUAL },
            { aType = HoJ.TYPE_RESULT, aConst = 1, bType = HoJ.TYPE_RESULT, bConst = 2, op = HoJ.OP_AND },
            { aType = HoJ.TYPE_INCOMING_MAJOR_DAMAGE_SPELL, aConst = 0, bType = HoJ.TYPE_CONSTANT, bConst = true, op = HoJ.OP_EQUAL },
            { aType = HoJ.TYPE_NET_INCOMING_DPS_PCT, aConst = 1.0, bType = HoJ.TYPE_CONSTANT, bConst = 0.25, op = HoJ.OP_GREATER_EQUAL },
            { aType = HoJ.TYPE_RESULT, aConst = 4, bType = HoJ.TYPE_RESULT, bConst = 5, op = HoJ.OP_OR },
            { aType = HoJ.TYPE_RESULT, aConst = 3, bType = HoJ.TYPE_RESULT, bConst = 6, op = HoJ.OP_AND },
        }
    },
    {
        spellInfo = paladinSpellInfo.blessingOfFreedom,
        -- _debugName = "blessingOfFreedom",
        tests = {
            { aType = HoJ.TYPE_UNIT_IS_SLOWED, aConst = "player", bType = HoJ.TYPE_CONSTANT, bConst = true, op = HoJ.OP_EQUAL },
        }
    },
}

local protDpsCooldownRules = {
    _name = "protDpsCooldownRules",
    {
        spellInfo = paladinSpellInfo.momentOfGlory,
        -- _debugName = "momentOfGlory",
        tests = {
            { aType = HoJ.TYPE_SKILL_OFF_COOLDOWN, aConst = paladinSpellInfo.avengersShield, bType = HoJ.TYPE_CONSTANT, bConst = false, op = HoJ.OP_EQUAL },
        }
    },
    {
        spellInfo = paladinSpellInfo.avengingWrath,
        -- _debugName = "avengingWrath",
        tests = {
            { aType = HoJ.TYPE_COMBAT_TARGET_TIME_TO_LIVE, aConst = 1, bType = HoJ.TYPE_SPELL_DURATION, bConst = paladinSpellInfo.avengingWrath, op = HoJ.OP_GREATER },
            { aType = HoJ.TYPE_COMBAT_ATTACKS_HIT, aConst = 0, bType = HoJ.TYPE_CONSTANT, bConst = 1, op = HoJ.OP_GREATER },
            { aType = HoJ.TYPE_RESULT, aConst = 1, bType = HoJ.TYPE_RESULT, bConst = 2, op = HoJ.OP_AND },
            { aType = HoJ.TYPE_TARGET_IS_TOUGH, aConst = 0, bType = HoJ.TYPE_CONSTANT, bConst = true, op = HoJ.OP_EQUAL },
            { aType = HoJ.TYPE_RESULT, aConst = 3, bType = HoJ.TYPE_RESULT, bConst = 4, op = HoJ.OP_OR },
            { aType = HoJ.TYPE_PLAYER_BUFF_ACTIVE, aConst = paladinSpellInfo.avengingWrath.id, bType = HoJ.TYPE_CONSTANT, bConst = false, op = HoJ.OP_EQUAL },
            { aType = HoJ.TYPE_RESULT, aConst = 5, bType = HoJ.TYPE_RESULT, bConst = 6, op = HoJ.OP_AND },
        }
    },
}

-- every possible spell that can show on the cooldown bar must have a priorty in this list
local protCooldownSortPriorities = {
    _name = "protCooldownSortPriorities",
    [paladinSpellInfo.layOnHands.id] = 17,
    [paladinSpellInfo.wordOfGlory.id] = 16,
    [paladinSpellInfo.divineShield.id] = 15,
    [paladinSpellInfo.blessingOfProtection.id] = 14,
    [paladinSpellInfo.guardianOfAncientKings.id] = 13,
    [paladinSpellInfo.rebuke.id] = 12,
    [paladinSpellInfo.repentance.id] = 11,
    [paladinSpellInfo.shieldOfTheRighteous.id] = 10,
    [paladinSpellInfo.ardentDefender.id] = 9,
    [paladinSpellInfo.avengersShield.id] = 8,
    [paladinSpellInfo.fistOfJustice.id] = 7,
    [paladinSpellInfo.hammerOfJustice.id] = 6,
    [paladinSpellInfo.blindingLight.id] = 5,
    [paladinSpellInfo.momentOfGlory.id] = 4,
    [paladinSpellInfo.avengingWrath.id] = 3,
    [paladinSpellInfo.seraphim.id] = 2,
    [paladinSpellInfo.blessingOfFreedom.id] = 1,
}

local protActiveBuffList = {
    -- SLREMOVE
--[[
    [paladinSpellInfo.handOfTheProtector.id] = { isActive = false, start = 0, duration = 0, timeRemaining = 0, stacks = 0 },
    [paladinSpellInfo.lightOfTheProtector.id] = { isActive = false, start = 0, duration = 0, timeRemaining = 0, stacks = 0 },
--]]
    [paladinSpellInfo.shieldOfTheRighteousBuff.id] = { isActive = false, start = 0, duration = 0, timeRemaining = 0, stacks = 0 },
    [paladinSpellInfo.giftOfTheNaaru.id] = { isActive = false, start = 0, duration = 0, timeRemaining = 0, stacks = 0 },
    [paladinSpellInfo.ardentDefender.id] = { isActive = false, start = 0, duration = 0, timeRemaining = 0, stacks = 0 },
    [paladinSpellInfo.guardianOfAncientKings.id] = { isActive = false, start = 0, duration = 0, timeRemaining = 0, stacks = 0 },
    [paladinSpellInfo.layOnHands.id] = { isActive = false, start = 0, duration = 0, timeRemaining = 0, stacks = 0 },
    [paladinSpellInfo.blessingOfProtection.id] = { isActive = false, start = 0, duration = 0, timeRemaining = 0, stacks = 0 },
    [paladinSpellInfo.divineShield.id] = { isActive = false, start = 0, duration = 0, timeRemaining = 0, stacks = 0 },
    [paladinSpellInfo.forbearance.id] = { isActive = false, start = 0, duration = 0, timeRemaining = 0, stacks = 0 },
    [paladinSpellInfo.avengingWrath.id] = { isActive = false, start = 0, duration = 0, timeRemaining = 0, stacks = 0 },
    [paladinSpellInfo.protConsecrationBuff.id] = { isActive = false, start = 0, duration = 0, timeRemaining = 0, stacks = 0 },
    [paladinSpellInfo.avengersValor.id] = { isActive = false, start = 0, duration = 0, timeRemaining = 0, stacks = 0 },
    [paladinSpellInfo.divinePurposeBuff.id] = { isActive = false, start = 0, duration = 0, timeRemaining = 0, stacks = 0 },
    [paladinSpellInfo.shiningLightBuff.id] = { isActive = false, start = 0, duration = 0, timeRemaining = 0, stacks = 0 },
}

local protProcdBuffList = {
    { _spellInfo = paladinSpellInfo.shieldOfTheRighteousBuff },
    { _spellInfo = paladinSpellInfo.protConsecrationBuff },
    { _spellInfo = paladinSpellInfo.divinePurposeBuff },
    { _spellInfo = paladinSpellInfo.shiningLightBuff },
}

local protBuffSortPriorities = {
    _name = "protBuffSortPriorities",
    [paladinSpellInfo.shieldOfTheRighteousBuff.id] = 4,
    [paladinSpellInfo.shiningLightBuff.id] = 3,
    [paladinSpellInfo.divinePurposeBuff.id] = 2,
    [paladinSpellInfo.protConsecrationBuff.id] = 1,
}

local holyProcdBuffList = {
    { _spellInfo = nil }, --paladinSpellInfo.
}

-- AoE spell tables are used to track cast of AoE spells that cause buffs without
-- timing information. The spell table reference the buff table and the buff table is
-- used to store the timing info for a buff when the spell is cast, and to look up
-- the timing info when active buffs are checked.
local protAoeSpellTable = {
    [paladinSpellInfo.consecration.id] = {
        _buffId = paladinSpellInfo.protConsecrationBuff.id
    }
}

local protAoeBuffTable = {
    [paladinSpellInfo.protConsecrationBuff.id] = {
        _castById = paladinSpellInfo.consecration,
        _spellInfo = paladinSpellInfo.protConsecrationBuff,
        _startTime = 0.0,
        _duration = 0.0,
        _expiration = 0.0,
    }
}

HoJ[HoJ.CLASS_NAME_PALADIN] = {
    _spellInfos = HoJ._paladinSpellInfos,
    _className = "PALADIN",
    _defaultSpecId = HoJ.SPEC_PALADIN_RETRIBUTION,
    _powerColor = { r = 1.0, g = 1.0, b = 0.0, a = 1.0 },
    [HoJ.SPEC_NAME_PALADIN_HOLY] = {
        -- properties
        _name = HoJ._text["Holy"],
        _spec = HoJ.SPEC_PALADIN_HOLY,
        _role = HoJ.ROLE_HEALING,
        _enabled = false,    -- addon enabled for this spec?
        _multiTargetRange = 8,
        _minorHealHealthRatio = 0.9,
        _majorHealHealthRatio = 0.6,
        _defaultPowerPerSec = 3,
        _autoPowerGen = false,
        _powerType = 9,
        _powerGainStep = 0,
        _powerMax = 5,
        _usesPowerPooling = false,
        -- functions
        _updateFn = HoJ.Paladin_Update_Holy,
        _updateRenderFn = HoJ.Paladin_UpdateRender,
        _selectSingleOrMultiTargetFn = HoJ.Paladin_SelectSingleOrMultitarget,
        _resetFn = HoJ.Paladin_Reset,
        _nextSkillSelectFn = HoJ.NextSkillOffCooldown,
        _checkBlessingsFn = nil,
        -- tables
        _meleeAoESpells = holyMeleeAoESpells,
        _activeBuffList = nil,
        _activeBlessings = nil,
        _procdBuffList = holyProcdBuffList,
        _rotationRules = { _singleTarget = nil, _multiTarget = nil },
        _interruptRules = { _singleTarget = nil, _multiTarget = nil },
        _mitigationRules = nil,
        _dpsCooldownRules = nil,
        _cooldownSortPriorities = nil,
        _buffSortPriorities = nil,
        _aoeSpellTable = nil,
        _aoeBuffTable = nil,
        _friendlyRangeSpells = {
            [HoJ.RANGE_MELEE] = nil,
            [HoJ.RANGE_8_YARDS] = nil,
            [HoJ.RANGE_10_YARDS] = nil,
            [HoJ.RANGE_12_YARDS] = nil,
            [HoJ.RANGE_15_YARDS] = nil,
            [HoJ.RANGE_20_YARDS] = nil,
            [HoJ.RANGE_25_YARDS] = nil,
            [HoJ.RANGE_30_YARDS] = nil,
            [HoJ.RANGE_40_YARDS] = {
                _spellInfo = paladinSpellInfo.flashOfLight
            },
        },
        _enemyRangeSpells = {
            [HoJ.RANGE_MELEE] = {
                _spellInfo = paladinSpellInfo.crusaderStrike
            },
            [HoJ.RANGE_8_YARDS] = {
                _spellInfo = paladinSpellInfo.crusaderStrike
            },
            [HoJ.RANGE_10_YARDS] = {
                _spellInfo = paladinSpellInfo.hammerOfJustice
            },
            [HoJ.RANGE_12_YARDS] = {
                _spellInfo = paladinSpellInfo.hammerOfJustice
            },
            [HoJ.RANGE_15_YARDS] = {
                _spellInfo = paladinSpellInfo.hammerOfJustice
            },
            [HoJ.RANGE_20_YARDS] = {
                _spellInfo = paladinSpellInfo.hammerOfJustice
            },
            [HoJ.RANGE_25_YARDS] = nil, -- will be more like 28 yards
            [HoJ.RANGE_30_YARDS] = {
                _spellInfo = paladinSpellInfo.judgment
            },
            [HoJ.RANGE_40_YARDS] = {
                _spellInfo = paladinSpellInfo.judgment
            },
        },
    },
    [HoJ.SPEC_NAME_PALADIN_PROTECTION] = {
        -- properties
        _name = HoJ._text["Protection"],
        _spec = HoJ.SPEC_PALADIN_PROTECTION,
        _role = HoJ.ROLE_TANK,
        _enabled = true,    -- addon enabled for this spec?
        _multiTargetRange = 8, -- blessed hammer won't quite reach 8 yards, and only reaches its max distance directly behind the player
        _minorHealHealthRatio = 0.9,
        _majorHealHealthRatio = 0.6,
        _defaultPowerPerSec = 2,
        _autoPowerGen = false,
        _powerType = 9,
        _powerGainStep = 0,
        _powerMax = 5,
        _usesPowerPooling = false,
        -- functions
        _updateFn = HoJ.Paladin_Update,
        _updateRenderFn = HoJ.Paladin_UpdateRender,
        _resetFn = HoJ.Paladin_Reset,
        _selectSingleOrMultiTargetFn = HoJ.Paladin_SelectSingleOrMultitarget,
        _nextSkillSelectFn = HoJ.NextSkillOffCooldown,
        _checkBlessingsFn = nil,
        -- tables
        _meleeAoESpells = protMeleeAoESpells,
        _activeBuffList = protActiveBuffList,
        _procdBuffList = protProcdBuffList,
        _activeBlessings = nil,
        _rotationRules = { _singleTarget = protRotationRules, _multiTarget = protRotationRules },
        _interruptRules = { _singleTarget = protInterruptRules, _multiTarget = protInterruptRules },
        _mitigationRules = { _singleTarget = protMitigationRules, _multiTarget = protMitigationRules },
        _dpsCooldownRules = { _singleTarget = protDpsCooldownRules, _multiTarget = protDpsCooldownRules },
        _cooldownSortPriorities = protCooldownSortPriorities,
        _buffSortPriorities = protBuffSortPriorities,
        _aoeSpellTable = protAoeSpellTable,
        _aoeBuffTable = protAoeBuffTable,
        _friendlyRangeSpells = {
            [HoJ.RANGE_MELEE] = nil,
            [HoJ.RANGE_8_YARDS] = nil,
            [HoJ.RANGE_10_YARDS] = nil,
            [HoJ.RANGE_12_YARDS] = nil,
            [HoJ.RANGE_15_YARDS] = nil,
            [HoJ.RANGE_20_YARDS] = nil,
            [HoJ.RANGE_25_YARDS] = nil,
            [HoJ.RANGE_30_YARDS] = nil,
            [HoJ.RANGE_40_YARDS] = {
                _spellInfo = paladinSpellInfo.flashOfLight
            },
        },
        _enemyRangeSpells = {
            [HoJ.RANGE_MELEE] = {
                _spellInfo = paladinSpellInfo.rebuke
            },
            [HoJ.RANGE_8_YARDS] = {
                _spellInfo = paladinSpellInfo.rebuke
            },
            [HoJ.RANGE_10_YARDS] = {
                _spellInfo = paladinSpellInfo.hammerOfJustice
            },
            [HoJ.RANGE_12_YARDS] = {
                _spellInfo = paladinSpellInfo.hammerOfJustice
            },
            [HoJ.RANGE_15_YARDS] = {
                _spellInfo = paladinSpellInfo.hammerOfJustice
            },
            [HoJ.RANGE_20_YARDS] = {
                _spellInfo = paladinSpellInfo.hammerOfJustice
            },
            [HoJ.RANGE_25_YARDS] = nil, -- more like 28 yards due to CheckInteractDistance limitattions
            [HoJ.RANGE_30_YARDS] = {
                _spellInfo = paladinSpellInfo.judgment
            },
            [HoJ.RANGE_40_YARDS] = {
                _spellInfo = paladinSpellInfo.judgment
            },
        },
    },
    [HoJ.SPEC_NAME_PALADIN_RETRIBUTION] = {
        -- properties
        _name = HoJ._text["Retribution"],
        _spec = HoJ.SPEC_PALADIN_RETRIBUTION,
        _role = HoJ.ROLE_DPS,
        _enabled = true,    -- enabled for this spec?
        _multiTargetRange = 8, -- divine storm testing with range addons indicates < 8
        _minorHealHealthRatio = 0.9,
        _majorHealHealthRatio = 0.6,
        _defaultPowerPerSec = 2, -- test show 1.0 is quite good but this varies with haste
        _autoPowerGen = false,
        _powerType = 9,
        _powerGainStep = 0,
        _powerMax = 5,
        _usesPowerPooling = false,
        -- functions
        _updateFn = HoJ.Paladin_Update,
        _updateRenderFn = HoJ.Paladin_UpdateRender,
        _selectSingleOrMultiTargetFn = HoJ.Paladin_SelectSingleOrMultitarget,
        _resetFn = HoJ.Paladin_Reset,
        _nextSkillSelectFn = HoJ.NextSkillOffCooldown,
        _checkBlessingsFn = HoJ.Paladin_CheckBlessings,
        _damageEventHandlerFn = HoJ.Paladin_DamageEventHandler,
        _reportSessionCombatStatsFn = nil, -- HoJ.Paladin_ReportSessionCombatStats,
        -- tables
        _meleeAoESpells = retMeleeAoESpells,
        _activeBuffList = retActiveBuffList,
        _procdBuffList = retProcdBuffList,
        _activeBlessings = retActiveBlessingsList,
        _rotationRules = { _singleTarget = retRotationRules, _multiTarget = retRotationRules },
        _interruptRules = { _singleTarget = retInterruptRules, _multiTarget = retInterruptRules },
        _mitigationRules = { _singleTarget = retMitigationRules, _multiTarget = retMitigationRules },
        _dpsCooldownRules = { _singleTarget = retDpsCooldownRules, _multiTarget = retDpsCooldownRules },
        _cooldownSortPriorities = retCooldownSortPriorities,
        _buffSortPriorities = retBuffSortPriorities,
        _aoeSpellTable = retAoeSpellTable,
        _aoeBuffTable = retAoeBuffTable,
        _friendlyRangeSpells = {
            [HoJ.RANGE_MELEE] = nil,
            [HoJ.RANGE_8_YARDS] = nil,
            [HoJ.RANGE_10_YARDS] = nil,
            [HoJ.RANGE_12_YARDS] = nil,
            [HoJ.RANGE_15_YARDS] = nil,
            [HoJ.RANGE_20_YARDS] = nil,
            [HoJ.RANGE_25_YARDS] = nil,
            [HoJ.RANGE_30_YARDS] = nil,
            [HoJ.RANGE_40_YARDS] = {
                _spellInfo = paladinSpellInfo.flashOfLight
            },
        },
        _enemyRangeSpells = {
            [HoJ.RANGE_MELEE] = {
                _spellInfo = paladinSpellInfo.crusaderStrike
            },
            [HoJ.RANGE_8_YARDS] = {
                _spellInfo = paladinSpellInfo.crusaderStrike
            },
            [HoJ.RANGE_10_YARDS] = {
                _spellInfo = paladinSpellInfo.hammerOfJustice
            },
            [HoJ.RANGE_12_YARDS] = {
                _spellInfo = paladinSpellInfo.bladeOfJustice
            },
            [HoJ.RANGE_15_YARDS] = {
                _spellInfo = paladinSpellInfo.bladeOfJustice
            },
            [HoJ.RANGE_20_YARDS] = {
                _spellInfo = paladinSpellInfo.bladeOfJustice
            },
            [HoJ.RANGE_25_YARDS] = nil, -- more like 28 yards due to CheckInteractDistance limitations
            [HoJ.RANGE_30_YARDS] = {
                _spellInfo = paladinSpellInfo.judgment
            },
            [HoJ.RANGE_40_YARDS] = {
                _spellInfo = paladinSpellInfo.judgment
            },
        },
    }
}

-- -----------------------------------------------------------------------------
-- HoJ:LoadPaladin
-- -----------------------------------------------------------------------------
function HoJ:LoadPaladin()
    -- HoJ:Log( "LoadPaladin" )

    if ( HoJ._classId ~= HoJ.CLASS_ID_PALADIN ) then
        HoJ:Error( "LoadPaladin called but _classId is not paladin!" )
    end

    -- TOOD: all of these HoJ variables need to be generalized for each class
    HoJ.SpellInfo = HoJ._paladinSpellInfos

    HoJ._protActiveBuffList = protActiveBuffList
    HoJ._retActiveBuffList = retActiveBuffList

    HoJ._lowHealthMitigationList = lowHealthMitigationList
    HoJ._moderateMitigationList = moderateMitigationList
    HoJ._emergencyMitigationList = emergencyMitigationList

    HoJ._friendlyRangeSpells = friendlyRangeSpells    
    HoJ._enemyRangeSpells = enemyRangeSpells

    HoJ._retActiveBlessingsList = retActiveBlessingsList

    HoJ._combatAuraSpellInfo = HoJ.SpellInfo.retributionAura; -- spell info for the aura we want to check for in combat

    -- HoJ:Log( "_spellInfo is " .. tostring( HoJ[HoJ.CLASS_NAME_PALADIN][HoJ.SPEC_NAME_PALADIN_RETRIBUTION]._enemyRangeSpells[HoJ.RANGE_30_YARDS]._spellInfo ) )
--[[
    ----------------------------------------------------------------------------
    -- show paladin-specific UI
    for i, v in ipairs( HoJ._HoPo ) do
        --HoJ:DebugMsg( "Setting alpha " .. i );
        v:SetAlpha( 0.0 );
        HoJ._HoPoGlow[i]:SetAlpha( 0.0 )
    end

    -- hide tv frame
    HoJFrame.HPMaxed:SetAlpha( 0.0 );
    HoJFrame.HPMaxedGlow:SetAlpha( 0.0 );
--]]
end
