-- -----------------------------------------------------------------------------
-- -----------------------------------------------------------------------------
--  Hammer of Justas
--
-- Copyright (C) 2010-2020
-- By Jonathan E. Wright
--
-- FILE: HoJ_Rogue.lua
-- DESC: Rogue-specific functions and tables.

--[[
TODO:
- many rogue skills seem to rely on combos. For instance, Symbols of Death + Shadow Dance or
  Vanish + Shadowstrike. There's not really a great way of chaining these right nowwithout 
  ensuring two rules sync up.
- Pooling. Need to wait for energy to pool for skills that aren't chosen currently but should 
  be chosen next. We currently end up with Backstab appearing in these gaps when we should just 
  be waiting to pool energy.

BUGS:

NOTES:
--]]

local addonName, HoJ = ...

-- DEFAULT_CHAT_FRAME:AddMessage( "HoJ_Rogue.lua" )

local rogueSpellInfo = HoJ._rogueSpellInfos

-- -----------------------------------------------------------------------------
-- HoJ:Rogue_SelectSingleOrMultitarget
-- returns
--   rotationRules
--   interruptRules
--   mitigationRules
--   dpsCooldownRules
-- -----------------------------------------------------------------------------
function HoJ:Rogue_SelectSingleOrMultitarget( curSpecInfo, numAttackers, threshold )
    if ( self._numMeleeAttackers > threshold ) then
        return curSpecInfo._rotationRules._multiTarget, 
                curSpecInfo._interruptRules._multiTarget,
                curSpecInfo._mitigationRules._multiTarget,
                curSpecInfo._dpsCooldownRules._multiTarget
    else
        return curSpecInfo._rotationRules._singleTarget, 
                curSpecInfo._interruptRules._singleTarget,
                curSpecInfo._mitigationRules._singleTarget,
                curSpecInfo._dpsCooldownRules._singleTarget
    end
end

-- -----------------------------------------------------------------------------
-- HoJ:Rogue_Update
-- -----------------------------------------------------------------------------
function HoJ:Rogue_Update( curSpecInfo, updateCooldowns, updateBuffs, updateRotation )
    -- HoJ:DebugMsg( "Updating spec " .. specInfo._name );
    local rotationRules, interruptRules, mitigationRules, dpsCooldownRules = curSpecInfo._selectSingleOrMultiTargetFn( 
        HoJ, curSpecInfo,
        self._numMeleeAttackers, 1 )

    -- update buffs
    if ( updateBuffs ) then
        local startBuffsMs = HoJ.ProfileBegin2( "Buffs" )
        
        HoJ:ClearIconList( HoJ._buffs );

        HoJ:CheckProcdBuffs( curSpecInfo._activeBuffList, curSpecInfo._procdBuffList );
        --HoJ:Warrior_CheckBlessings( curSpecInfo._activeBlessings );

        HoJ.ProfileFunction( "UpdateIcons2:Buffs2:", HoJ.UpdateIcons2, self, HoJ._buffs, 
                curSpecInfo._buffSortPriorities, HoJ._buffFrames, true )

        HoJ.ProfileEnd( "Buffs", startBuffsMs )
    end

    -- update cooldowns
    if ( updateCooldowns ) then
        local startCooldownsMs = HoJ.ProfileBegin2( "Cooldowns" )

        HoJ.ProfileFunction( "ClearIconList", HoJ.ClearIconList, self, HoJ._cooldowns )

        HoJ.ProfileFunction( "CheckRules(dpsCooldowns)", HoJ.CheckRules, self, self._cooldowns,
                self._showCooldownsFrame, dpsCooldownRules, true, true )
        HoJ.ProfileFunction( "CheckRules(interrupts)", HoJ.CheckRules, self, self._cooldowns,
                self._showCooldownsFrame, interruptRules, true, true )
        HoJ.ProfileFunction( "CheckRules(mitigations)", HoJ.CheckRules, self, self._cooldowns, 
                self._showCooldownsFrame, mitigationRules, true, true )

        HoJ.ProfileFunction( "UpdateIcons2:Cooldowns:", HoJ.UpdateIcons2, self, self._cooldowns, 
                curSpecInfo._cooldownSortPriorities, self._cooldownFrames, true )

        HoJ.ProfileEnd( "Cooldowns", startCooldownsMs )
    end

    -- update rotation
    if ( updateRotation ) then
        local startRotationMs = HoJ.ProfileBegin2( "Rotation" )

        HoJ:ClearIconList( HoJ._rotationSkills )

        HoJ:CheckDPSRotation( HoJ._rotationSkills, rotationRules )

        HoJ.ProfileFunction( "UpdateIcons2:Rotation:", HoJ.UpdateIcons2, self, HoJ._rotationSkills, 
                { }, HoJ._rotationFrames, false );

        HoJ.ProfileEnd( "Rotation", startRotationMs )
    end    
end

--------------------------------------------------------------------------------
-- SUBTLETY
--------------------------------------------------------------------------------

-- https://www.icy-veins.com/wow/subtlety-rogue-pve-dps-rotation-cooldowns-abilities
local subtletyRotationRules = {
    _name = "subtletyRotationRules",
    {
        spellInfo = rogueSpellInfo.shadowBlades,
        _debugName = "shadowBlades",
        tests = {
            { aType = HoJ.TYPE_CONSTANT, aConst = true, bType = HoJ.TYPE_CONSTANT, bConst = true, op = HoJ.OP_AND },
        }
    },
    {
        -- nightblade if target will outlive the debuff
        spellInfo = rogueSpellInfo.nightblade,
        _debugName = "nightblade",
        tests = {
            { aType = HoJ.TYPE_TARGET_PLAYER_DEBUFF_REMAINING, aConst = rogueSpellInfo.nightbladeDebuff.id, bType = HoJ.TYPE_CONSTANT, bConst = 2.0, op = HoJ.OP_LESS_EQUAL },
            { aType = HoJ.TYPE_TARGET_RANGE, aConst = rogueSpellInfo.nightblade.range, bType = HoJ.TYPE_CONSTANT, bConst = true, op = HoJ.OP_EQUAL },
            { aType = HoJ.TYPE_RESULT, aConst = 1, bType = HoJ.TYPE_RESULT, bConst = 2, op = HoJ.OP_AND },
            { aType = HoJ.TYPE_COMBO_POINTS, aConst = 1, bType = HoJ.TYPE_CONSTANT, bConst = 4, op = HoJ.OP_GREATER_EQUAL },
            { aType = HoJ.TYPE_COMBAT_TARGET_TIME_TO_LIVE, aConst = 2, bType = HoJ.TYPE_BUFF_DURATION_FOR_COMBO_POINTS, bConst = rogueSpellInfo.nightbladeDebuff, op = HoJ.OP_GREATER_EQUAL },
            { aType = HoJ.TYPE_RESULT, aConst = 4, bType = HoJ.TYPE_RESULT, bConst = 5, op = HoJ.OP_AND },
            { aType = HoJ.TYPE_RESULT, aConst = 3, bType = HoJ.TYPE_RESULT, bConst = 6, op = HoJ.OP_AND },
        }
    },
    -- TODO: combos
    {
        -- symbols of death of off cooldown and we have > 50 power
        spellInfo = rogueSpellInfo.symbolsOfDeath,
         _debugName = "symbolsOfDeath",
        tests = {
            { aType = HoJ.TYPE_POWER, aConst = 1, bType = HoJ.TYPE_CONSTANT, bConst = 50, op = HoJ.OP_GREATER_EQUAL },
            { aType = HoJ.TYPE_PLAYER_IS_STEALTHED, aConst = 1, bType = HoJ.TYPE_CONSTANT, bConst = false, op = HoJ.OP_EQUAL },
            { aType = HoJ.TYPE_RESULT, aConst = 1, bType = HoJ.TYPE_RESULT, bConst = 2, op = HoJ.OP_AND },
        }
    },
    {
        -- shadow dance if the symbols of death buff is active and not in stealth
        spellInfo = rogueSpellInfo.shadowDance,
        _debugName = "shadowDance",
        tests = {
            { aType = HoJ.TYPE_PLAYER_BUFF_ACTIVE, aConst = rogueSpellInfo.symbolsOfDeath.id, bType = HoJ.TYPE_CONSTANT, bConst = true, op = HoJ.OP_EQUAL },
            { aType = HoJ.TYPE_PLAYER_BUFF_ACTIVE, aConst = rogueSpellInfo.shadowDanceBuff.id, bType = HoJ.TYPE_CONSTANT, bConst = false, op = HoJ.OP_EQUAL },
            { aType = HoJ.TYPE_RESULT, aConst = 1, bType = HoJ.TYPE_RESULT, bConst = 2, op = HoJ.OP_AND },
            { aType = HoJ.TYPE_PLAYER_IS_STEALTHED, aConst = 1, bType = HoJ.TYPE_CONSTANT, bConst = false, op = HoJ.OP_EQUAL },
            { aType = HoJ.TYPE_RESULT, aConst = 3, bType = HoJ.TYPE_RESULT, bConst = 4, op = HoJ.OP_AND },
        }
    },
    {
        -- marked for death if off cooldown and if shadow dance is not active
        spellInfo = rogueSpellInfo.markedForDeath,
        _debugName = "markedForDeath",
        tests = {
            { aType = HoJ.TYPE_PLAYER_BUFF_ACTIVE, aConst = rogueSpellInfo.shadowDanceBuff.id, bType = HoJ.TYPE_CONSTANT, bConst = false, op = HoJ.OP_EQUAL },
            { aType = HoJ.TYPE_TARGET_RANGE, aConst = rogueSpellInfo.markedForDeath.range, bType = HoJ.TYPE_CONSTANT, bConst = true, op = HoJ.OP_EQUAL },
            { aType = HoJ.TYPE_RESULT, aConst = 1, bType = HoJ.TYPE_RESULT, bConst = 2, op = HoJ.OP_AND },
        }
    },
    -- TODO: vanish + shadowstrike combo
    {
        -- secret technique if >= 4 combo points
        spellInfo = rogueSpellInfo.secretTechnique,
        _debugName = "secretTechnique_with_deeperStrategem",
        tests = {
            { aType = HoJ.TYPE_TALENT_SELECTED, aConst = rogueSpellInfo.deeperStrategem, bType = HoJ.TYPE_CONSTANT, bConst = true, op = HoJ.OP_EQUAL },
            { aType = HoJ.TYPE_COMBO_POINTS, aConst = 1, bType = HoJ.TYPE_CONSTANT, bConst = 5, op = HoJ.OP_GREATER_EQUAL },
            { aType = HoJ.TYPE_RESULT, aConst = 1, bType = HoJ.TYPE_RESULT, bConst = 2, op = HoJ.OP_AND },
            { aType = HoJ.TYPE_TARGET_RANGE, aConst = rogueSpellInfo.secretTechnique.range, bType = HoJ.TYPE_CONSTANT, bConst = true, op = HoJ.OP_EQUAL },
            { aType = HoJ.TYPE_RESULT, aConst = 3, bType = HoJ.TYPE_RESULT, bConst = 4, op = HoJ.OP_AND },
            { aType = HoJ.TYPE_PLAYER_BUFF_ACTIVE, aConst = rogueSpellInfo.nightbladeDebuff.id, bType = HoJ.TYPE_CONSTANT, bConst = true, op = HoJ.OP_EQUAL },
            { aType = HoJ.TYPE_TARGET_PLAYER_DEBUFF_REMAINING, aConst = rogueSpellInfo.nightbladeDebuff.id, bType = HoJ.TYPE_CONSTANT, bConst = 0.0, op = HoJ.OP_GREATER },
            { aType = HoJ.TYPE_RESULT, aConst = 6, bType = HoJ.TYPE_RESULT, bConst = 7, op = HoJ.OP_OR },
            { aType = HoJ.TYPE_COMBAT_TARGET_TIME_TO_LIVE, aConst = 2, bType = HoJ.TYPE_BUFF_DURATION_FOR_COMBO_POINTS, bConst = rogueSpellInfo.nightbladeDebuff, op = HoJ.OP_LESS },
            { aType = HoJ.TYPE_RESULT, aConst = 8, bType = HoJ.TYPE_RESULT, bConst = 9, op = HoJ.OP_OR },
            { aType = HoJ.TYPE_RESULT, aConst = 5, bType = HoJ.TYPE_RESULT, bConst = 10, op = HoJ.OP_AND },
        }
    },
    {
        -- eviscerate at 4+ combo points if not talented for Deeper Strategem
        spellInfo = rogueSpellInfo.secretTechnique,
        _debugName = "secretTechnique",
        tests = {
            { aType = HoJ.TYPE_TALENT_SELECTED, aConst = rogueSpellInfo.deeperStrategem, bType = HoJ.TYPE_CONSTANT, bConst = false, op = HoJ.OP_EQUAL },
            { aType = HoJ.TYPE_COMBO_POINTS, aConst = 1, bType = HoJ.TYPE_CONSTANT, bConst = 4, op = HoJ.OP_GREATER_EQUAL },
            { aType = HoJ.TYPE_RESULT, aConst = 1, bType = HoJ.TYPE_RESULT, bConst = 2, op = HoJ.OP_AND },
            { aType = HoJ.TYPE_TARGET_RANGE, aConst = rogueSpellInfo.secretTechnique.range, bType = HoJ.TYPE_CONSTANT, bConst = true, op = HoJ.OP_EQUAL },
            { aType = HoJ.TYPE_RESULT, aConst = 3, bType = HoJ.TYPE_RESULT, bConst = 4, op = HoJ.OP_AND },
            { aType = HoJ.TYPE_PLAYER_BUFF_ACTIVE, aConst = rogueSpellInfo.nightbladeDebuff.id, bType = HoJ.TYPE_CONSTANT, bConst = true, op = HoJ.OP_EQUAL },
            { aType = HoJ.TYPE_TARGET_PLAYER_DEBUFF_REMAINING, aConst = rogueSpellInfo.nightbladeDebuff.id, bType = HoJ.TYPE_CONSTANT, bConst = 0.0, op = HoJ.OP_GREATER },
            { aType = HoJ.TYPE_RESULT, aConst = 6, bType = HoJ.TYPE_RESULT, bConst = 7, op = HoJ.OP_OR },
            { aType = HoJ.TYPE_COMBAT_TARGET_TIME_TO_LIVE, aConst = 2, bType = HoJ.TYPE_BUFF_DURATION_FOR_COMBO_POINTS, bConst = rogueSpellInfo.nightbladeDebuff, op = HoJ.OP_LESS },
            { aType = HoJ.TYPE_RESULT, aConst = 8, bType = HoJ.TYPE_RESULT, bConst = 9, op = HoJ.OP_OR },
            { aType = HoJ.TYPE_RESULT, aConst = 5, bType = HoJ.TYPE_RESULT, bConst = 10, op = HoJ.OP_AND },
        }
    },
    {
        -- eviscerate at 5+ combo points if talented for Deeper Strategem
        spellInfo = rogueSpellInfo.eviscerate,
        _debugName = "eviscerate_with_deeperStrategem",
        tests = {
            { aType = HoJ.TYPE_TALENT_SELECTED, aConst = rogueSpellInfo.deeperStrategem, bType = HoJ.TYPE_CONSTANT, bConst = true, op = HoJ.OP_EQUAL },
            { aType = HoJ.TYPE_COMBO_POINTS, aConst = 1, bType = HoJ.TYPE_CONSTANT, bConst = 5, op = HoJ.OP_GREATER_EQUAL },
            { aType = HoJ.TYPE_RESULT, aConst = 1, bType = HoJ.TYPE_RESULT, bConst = 2, op = HoJ.OP_AND },
            { aType = HoJ.TYPE_TARGET_RANGE, aConst = rogueSpellInfo.eviscerate.range, bType = HoJ.TYPE_CONSTANT, bConst = true, op = HoJ.OP_EQUAL },
            { aType = HoJ.TYPE_RESULT, aConst = 3, bType = HoJ.TYPE_RESULT, bConst = 4, op = HoJ.OP_AND },
            { aType = HoJ.TYPE_PLAYER_BUFF_ACTIVE, aConst = rogueSpellInfo.nightbladeDebuff.id, bType = HoJ.TYPE_CONSTANT, bConst = true, op = HoJ.OP_EQUAL },
            { aType = HoJ.TYPE_TARGET_PLAYER_DEBUFF_REMAINING, aConst = rogueSpellInfo.nightbladeDebuff.id, bType = HoJ.TYPE_CONSTANT, bConst = 0.0, op = HoJ.OP_GREATER },
            { aType = HoJ.TYPE_RESULT, aConst = 6, bType = HoJ.TYPE_RESULT, bConst = 7, op = HoJ.OP_OR },
            { aType = HoJ.TYPE_COMBAT_TARGET_TIME_TO_LIVE, aConst = 2, bType = HoJ.TYPE_BUFF_DURATION_FOR_COMBO_POINTS, bConst = rogueSpellInfo.nightbladeDebuff, op = HoJ.OP_LESS },
            { aType = HoJ.TYPE_RESULT, aConst = 8, bType = HoJ.TYPE_RESULT, bConst = 9, op = HoJ.OP_OR },
            { aType = HoJ.TYPE_RESULT, aConst = 5, bType = HoJ.TYPE_RESULT, bConst = 10, op = HoJ.OP_AND },
        }
    },
    {
        -- eviscerate at 4+ combo points if not talented for Deeper Strategem
        spellInfo = rogueSpellInfo.eviscerate,
        _debugName = "eviscerate",
        tests = {
            { aType = HoJ.TYPE_TALENT_SELECTED, aConst = rogueSpellInfo.deeperStrategem, bType = HoJ.TYPE_CONSTANT, bConst = false, op = HoJ.OP_EQUAL },
            { aType = HoJ.TYPE_COMBO_POINTS, aConst = 1, bType = HoJ.TYPE_CONSTANT, bConst = 4, op = HoJ.OP_GREATER_EQUAL },
            { aType = HoJ.TYPE_RESULT, aConst = 1, bType = HoJ.TYPE_RESULT, bConst = 2, op = HoJ.OP_AND },
            { aType = HoJ.TYPE_TARGET_RANGE, aConst = rogueSpellInfo.eviscerate.range, bType = HoJ.TYPE_CONSTANT, bConst = true, op = HoJ.OP_EQUAL },
            { aType = HoJ.TYPE_RESULT, aConst = 3, bType = HoJ.TYPE_RESULT, bConst = 4, op = HoJ.OP_AND },
            { aType = HoJ.TYPE_PLAYER_BUFF_ACTIVE, aConst = rogueSpellInfo.nightbladeDebuff.id, bType = HoJ.TYPE_CONSTANT, bConst = true, op = HoJ.OP_EQUAL },
            { aType = HoJ.TYPE_TARGET_PLAYER_DEBUFF_REMAINING, aConst = rogueSpellInfo.nightbladeDebuff.id, bType = HoJ.TYPE_CONSTANT, bConst = 0.0, op = HoJ.OP_GREATER },
            { aType = HoJ.TYPE_RESULT, aConst = 6, bType = HoJ.TYPE_RESULT, bConst = 7, op = HoJ.OP_OR },
            { aType = HoJ.TYPE_COMBAT_TARGET_TIME_TO_LIVE, aConst = 2, bType = HoJ.TYPE_BUFF_DURATION_FOR_COMBO_POINTS, bConst = rogueSpellInfo.nightbladeDebuff, op = HoJ.OP_LESS },
            { aType = HoJ.TYPE_RESULT, aConst = 8, bType = HoJ.TYPE_RESULT, bConst = 9, op = HoJ.OP_OR },
            { aType = HoJ.TYPE_RESULT, aConst = 5, bType = HoJ.TYPE_RESULT, bConst = 10, op = HoJ.OP_AND },
        }
    },
    {
        spellInfo = rogueSpellInfo.shadowstrike,
        _debugName = "shadowstrike",
        tests = {
            { aType = HoJ.TYPE_PLAYER_IS_STEALTHED, aConst = 1, bType = HoJ.TYPE_CONSTANT, bConst = true, op = HoJ.OP_EQUAL },
            { aType = HoJ.TYPE_PLAYER_BUFF_ACTIVE, aConst = rogueSpellInfo.shadowDanceBuff.id, bType = HoJ.TYPE_CONSTANT, bConst = true, op = HoJ.OP_EQUAL },
            { aType = HoJ.TYPE_RESULT, aConst = 1, bType = HoJ.TYPE_RESULT, bConst = 2, op = HoJ.OP_OR },
            { aType = HoJ.TYPE_TARGET_RANGE, aConst = rogueSpellInfo.shadowstrike.range, bType = HoJ.TYPE_CONSTANT, bConst = true, op = HoJ.OP_EQUAL },
            { aType = HoJ.TYPE_RESULT, aConst = 3, bType = HoJ.TYPE_RESULT, bConst = 4, op = HoJ.OP_AND },
        }
    },
    {
        -- backstab as long we will still have > 50 energy when symbols of death is off cooldown
        spellInfo = rogueSpellInfo.gloomblade,
        _debugName = "gloomblade",
        tests = {
            { 
                aType = HoJ.TYPE_POWER_AFTER_CD_MINUS_SKILL_POWER,
                aConst = rogueSpellInfo.symbolsOfDeath.id,
                aConst2 = rogueSpellInfo.gloomblade,
                bType = HoJ.TYPE_CONSTANT,
                bConst = 50,
                op = HoJ.OP_GREATER_EQUAL
            },
            { aType = HoJ.TYPE_TARGET_RANGE, aConst = rogueSpellInfo.gloomblade.range, bType = HoJ.TYPE_CONSTANT, bConst = true, op = HoJ.OP_EQUAL },
            { aType = HoJ.TYPE_RESULT, aConst = 1, bType = HoJ.TYPE_RESULT, bConst = 2, op = HoJ.OP_AND },
        }
    },
    {
        -- backstab as long we will still have > 50 energy when symbols of death is off cooldown
        spellInfo = rogueSpellInfo.backstab,
        replacedByTalent = rogueSpellInfo.gloomblade,
        _debugName = "backstab",
        -- _debugFail = true,
        tests = {
            { 
                aType = HoJ.TYPE_POWER_AFTER_CD_MINUS_SKILL_POWER,
                aConst = rogueSpellInfo.symbolsOfDeath.id,
                aConst2 = rogueSpellInfo.backstab,
                bType = HoJ.TYPE_CONSTANT,
                bConst = 50,
                op = HoJ.OP_GREATER_EQUAL
            },
            { aType = HoJ.TYPE_TARGET_RANGE, aConst = rogueSpellInfo.backstab.range, bType = HoJ.TYPE_CONSTANT, bConst = true, op = HoJ.OP_EQUAL },
            { aType = HoJ.TYPE_RESULT, aConst = 1, bType = HoJ.TYPE_RESULT, bConst = 2, op = HoJ.OP_AND },
        }
    },
}

local subtletyMultiTargetRotationRules = {
    _name = "subtletyMultitargetRotationRules",
    {
        -- symbols of death of off cooldown and we have > 50 power
        spellInfo = rogueSpellInfo.symbolsOfDeath,
         _debugName = "symbolsOfDeath",
        tests = {
            { aType = HoJ.TYPE_POWER, aConst = 1, bType = HoJ.TYPE_CONSTANT, bConst = 50, op = HoJ.OP_GREATER_EQUAL },
            { aType = HoJ.TYPE_PLAYER_IS_STEALTHED, aConst = 1, bType = HoJ.TYPE_CONSTANT, bConst = false, op = HoJ.OP_EQUAL },
            { aType = HoJ.TYPE_RESULT, aConst = 1, bType = HoJ.TYPE_RESULT, bConst = 2, op = HoJ.OP_AND },
        }
    },
    {
        -- shadow dance if the symbols of death buff is active
        spellInfo = rogueSpellInfo.shadowDance,
        _debugName = "shadowDance",
        tests = {
            { aType = HoJ.TYPE_PLAYER_BUFF_ACTIVE, aConst = rogueSpellInfo.symbolsOfDeath.id, bType = HoJ.TYPE_CONSTANT, bConst = true, op = HoJ.OP_EQUAL },
            { aType = HoJ.TYPE_PLAYER_BUFF_ACTIVE, aConst = rogueSpellInfo.shadowDanceBuff.id, bType = HoJ.TYPE_CONSTANT, bConst = false, op = HoJ.OP_EQUAL },
            { aType = HoJ.TYPE_RESULT, aConst = 1, bType = HoJ.TYPE_RESULT, bConst = 2, op = HoJ.OP_AND },
            { aType = HoJ.TYPE_PLAYER_IS_STEALTHED, aConst = 1, bType = HoJ.TYPE_CONSTANT, bConst = false, op = HoJ.OP_EQUAL },
            { aType = HoJ.TYPE_RESULT, aConst = 3, bType = HoJ.TYPE_RESULT, bConst = 4, op = HoJ.OP_AND },
        }
    },
    {
        -- nightblade if target will outlive the debuff
        spellInfo = rogueSpellInfo.nightblade,
        _debugName = "nightblade",
        tests = {
            { aType = HoJ.TYPE_TARGET_PLAYER_DEBUFF_REMAINING, aConst = rogueSpellInfo.nightbladeDebuff.id, bType = HoJ.TYPE_CONSTANT, bConst = 2.0, op = HoJ.OP_LESS_EQUAL },
            { aType = HoJ.TYPE_TARGET_RANGE, aConst = rogueSpellInfo.nightblade.range, bType = HoJ.TYPE_CONSTANT, bConst = true, op = HoJ.OP_EQUAL },
            { aType = HoJ.TYPE_RESULT, aConst = 1, bType = HoJ.TYPE_RESULT, bConst = 2, op = HoJ.OP_AND },
            { aType = HoJ.TYPE_COMBO_POINTS, aConst = 1, bType = HoJ.TYPE_CONSTANT, bConst = 4, op = HoJ.OP_GREATER_EQUAL },
            { aType = HoJ.TYPE_COMBAT_TARGET_TIME_TO_LIVE, aConst = 2, bType = HoJ.TYPE_BUFF_DURATION_FOR_COMBO_POINTS, bConst = rogueSpellInfo.nightbladeDebuff, op = HoJ.OP_GREATER_EQUAL },
            { aType = HoJ.TYPE_RESULT, aConst = 4, bType = HoJ.TYPE_RESULT, bConst = 5, op = HoJ.OP_AND },
            { aType = HoJ.TYPE_RESULT, aConst = 3, bType = HoJ.TYPE_RESULT, bConst = 6, op = HoJ.OP_AND },
        }
    },

    {
        -- secret technique if >= 4 combo points
        spellInfo = rogueSpellInfo.secretTechnique,
        _debugName = "secretTechnique",
        tests = {
            { aType = HoJ.TYPE_TARGET_RANGE, aConst = rogueSpellInfo.secretTechnique.range, bType = HoJ.TYPE_CONSTANT, bConst = true, op = HoJ.OP_EQUAL },
            { aType = HoJ.TYPE_COMBO_POINTS, aConst = 1, bType = HoJ.TYPE_CONSTANT, bConst = 4, op = HoJ.OP_GREATER_EQUAL },
            { aType = HoJ.TYPE_RESULT, aConst = 1, bType = HoJ.TYPE_RESULT, bConst = 2, op = HoJ.OP_AND },
        }
    },
    {
        -- marked for death if off cooldown and if shadow dance is not active
        spellInfo = rogueSpellInfo.markedForDeath,
        _debugName = "markedForDeath",
        tests = {
            { aType = HoJ.TYPE_PLAYER_BUFF_ACTIVE, aConst = rogueSpellInfo.shadowDanceBuff.id, bType = HoJ.TYPE_CONSTANT, bConst = false, op = HoJ.OP_EQUAL },
            { aType = HoJ.TYPE_TARGET_RANGE, aConst = rogueSpellInfo.markedForDeath.range, bType = HoJ.TYPE_CONSTANT, bConst = true, op = HoJ.OP_EQUAL },
            { aType = HoJ.TYPE_RESULT, aConst = 1, bType = HoJ.TYPE_RESULT, bConst = 2, op = HoJ.OP_AND },
        }
    },
    {
        spellInfo = rogueSpellInfo.shurikenStorm,
        _debugName = "shurikenStorm",
        tests = {
            { aType = HoJ.TYPE_PLAYER_BUFF_ACTIVE, aConst = rogueSpellInfo.shadowDanceBuff.id, bType = HoJ.TYPE_CONSTANT, bConst = true, op = HoJ.OP_EQUAL },
            { aType = HoJ.TYPE_NUM_MELEE_ATTACKERS, aConst = 1, bType = HoJ.TYPE_CONSTANT, bConst = 3, op = HoJ.OP_GREATER_EQUAL },
            { aType = HoJ.TYPE_RESULT, aConst = 1, bType = HoJ.TYPE_RESULT, bConst = 2, op = HoJ.OP_OR },
            { aType = HoJ.TYPE_COMBO_POINTS, aConst = 1, bType = HoJ.TYPE_CONSTANT, bConst = 4, op = HoJ.OP_LESS },
            { aType = HoJ.TYPE_TARGET_RANGE, aConst = rogueSpellInfo.shurikenStorm.range, bType = HoJ.TYPE_CONSTANT, bConst = true, op = HoJ.OP_EQUAL },
            { aType = HoJ.TYPE_RESULT, aConst = 4, bType = HoJ.TYPE_RESULT, bConst = 5, op = HoJ.OP_AND },
            { aType = HoJ.TYPE_RESULT, aConst = 3, bType = HoJ.TYPE_RESULT, bConst = 6, op = HoJ.OP_AND },
        }
    },
    {
        -- secretTechnique at 5+ combo points if talented for Deeper Strategem
        spellInfo = rogueSpellInfo.secretTechnique,
        _debugName = "secretTechnique_with_deeperStrategem",
        tests = {
            { aType = HoJ.TYPE_TALENT_SELECTED, aConst = rogueSpellInfo.deeperStrategem, bType = HoJ.TYPE_CONSTANT, bConst = true, op = HoJ.OP_EQUAL },
            { aType = HoJ.TYPE_COMBO_POINTS, aConst = 1, bType = HoJ.TYPE_CONSTANT, bConst = 5, op = HoJ.OP_GREATER_EQUAL },
            { aType = HoJ.TYPE_RESULT, aConst = 1, bType = HoJ.TYPE_RESULT, bConst = 2, op = HoJ.OP_AND },
            { aType = HoJ.TYPE_TARGET_RANGE, aConst = rogueSpellInfo.secretTechnique.range, bType = HoJ.TYPE_CONSTANT, bConst = true, op = HoJ.OP_EQUAL },
            { aType = HoJ.TYPE_RESULT, aConst = 3, bType = HoJ.TYPE_RESULT, bConst = 4, op = HoJ.OP_AND },
            { aType = HoJ.TYPE_PLAYER_BUFF_ACTIVE, aConst = rogueSpellInfo.nightbladeDebuff.id, bType = HoJ.TYPE_CONSTANT, bConst = true, op = HoJ.OP_EQUAL },
            { aType = HoJ.TYPE_TARGET_PLAYER_DEBUFF_REMAINING, aConst = rogueSpellInfo.nightbladeDebuff.id, bType = HoJ.TYPE_CONSTANT, bConst = 0.0, op = HoJ.OP_GREATER },
            { aType = HoJ.TYPE_RESULT, aConst = 6, bType = HoJ.TYPE_RESULT, bConst = 7, op = HoJ.OP_OR },
            { aType = HoJ.TYPE_COMBAT_TARGET_TIME_TO_LIVE, aConst = 2, bType = HoJ.TYPE_BUFF_DURATION_FOR_COMBO_POINTS, bConst = rogueSpellInfo.nightbladeDebuff, op = HoJ.OP_LESS },
            { aType = HoJ.TYPE_RESULT, aConst = 8, bType = HoJ.TYPE_RESULT, bConst = 9, op = HoJ.OP_OR },
            { aType = HoJ.TYPE_RESULT, aConst = 5, bType = HoJ.TYPE_RESULT, bConst = 10, op = HoJ.OP_AND },
        }
    },
    {
        -- secretTechnique at 4+ combo points if not talented for Deeper Strategem
        spellInfo = rogueSpellInfo.secretTechnique,
        _debugName = "secretTechnique",
        tests = {
            { aType = HoJ.TYPE_TALENT_SELECTED, aConst = rogueSpellInfo.deeperStrategem, bType = HoJ.TYPE_CONSTANT, bConst = false, op = HoJ.OP_EQUAL },
            { aType = HoJ.TYPE_COMBO_POINTS, aConst = 1, bType = HoJ.TYPE_CONSTANT, bConst = 4, op = HoJ.OP_GREATER_EQUAL },
            { aType = HoJ.TYPE_RESULT, aConst = 1, bType = HoJ.TYPE_RESULT, bConst = 2, op = HoJ.OP_AND },
            { aType = HoJ.TYPE_TARGET_RANGE, aConst = rogueSpellInfo.secretTechnique.range, bType = HoJ.TYPE_CONSTANT, bConst = true, op = HoJ.OP_EQUAL },
            { aType = HoJ.TYPE_RESULT, aConst = 3, bType = HoJ.TYPE_RESULT, bConst = 4, op = HoJ.OP_AND },
            { aType = HoJ.TYPE_PLAYER_BUFF_ACTIVE, aConst = rogueSpellInfo.nightbladeDebuff.id, bType = HoJ.TYPE_CONSTANT, bConst = true, op = HoJ.OP_EQUAL },
            { aType = HoJ.TYPE_TARGET_PLAYER_DEBUFF_REMAINING, aConst = rogueSpellInfo.nightbladeDebuff.id, bType = HoJ.TYPE_CONSTANT, bConst = 0.0, op = HoJ.OP_GREATER },
            { aType = HoJ.TYPE_RESULT, aConst = 6, bType = HoJ.TYPE_RESULT, bConst = 7, op = HoJ.OP_OR },
            { aType = HoJ.TYPE_COMBAT_TARGET_TIME_TO_LIVE, aConst = 2, bType = HoJ.TYPE_BUFF_DURATION_FOR_COMBO_POINTS, bConst = rogueSpellInfo.nightbladeDebuff, op = HoJ.OP_LESS },
            { aType = HoJ.TYPE_RESULT, aConst = 8, bType = HoJ.TYPE_RESULT, bConst = 9, op = HoJ.OP_OR },
            { aType = HoJ.TYPE_RESULT, aConst = 5, bType = HoJ.TYPE_RESULT, bConst = 10, op = HoJ.OP_AND },
        }
    },
    {
        -- eviscerate at 5+ combo points if talented for Deeper Strategem
        spellInfo = rogueSpellInfo.eviscerate,
        _debugName = "eviscerate_with_deeperStrategem",
        tests = {
            { aType = HoJ.TYPE_TALENT_SELECTED, aConst = rogueSpellInfo.deeperStrategem, bType = HoJ.TYPE_CONSTANT, bConst = true, op = HoJ.OP_EQUAL },
            { aType = HoJ.TYPE_COMBO_POINTS, aConst = 1, bType = HoJ.TYPE_CONSTANT, bConst = 5, op = HoJ.OP_GREATER_EQUAL },
            { aType = HoJ.TYPE_RESULT, aConst = 1, bType = HoJ.TYPE_RESULT, bConst = 2, op = HoJ.OP_AND },
            { aType = HoJ.TYPE_TARGET_RANGE, aConst = rogueSpellInfo.eviscerate.range, bType = HoJ.TYPE_CONSTANT, bConst = true, op = HoJ.OP_EQUAL },
            { aType = HoJ.TYPE_RESULT, aConst = 3, bType = HoJ.TYPE_RESULT, bConst = 4, op = HoJ.OP_AND },
            { aType = HoJ.TYPE_PLAYER_BUFF_ACTIVE, aConst = rogueSpellInfo.nightblade.id, bType = HoJ.TYPE_CONSTANT, bConst = true, op = HoJ.OP_EQUAL },
            { aType = HoJ.TYPE_TARGET_PLAYER_DEBUFF_REMAINING, aConst = rogueSpellInfo.nightbladeDebuff.id, bType = HoJ.TYPE_CONSTANT, bConst = 0.0, op = HoJ.OP_GREATER },
            { aType = HoJ.TYPE_RESULT, aConst = 6, bType = HoJ.TYPE_RESULT, bConst = 7, op = HoJ.OP_OR },
            { aType = HoJ.TYPE_COMBAT_TARGET_TIME_TO_LIVE, aConst = 2, bType = HoJ.TYPE_BUFF_DURATION_FOR_COMBO_POINTS, bConst = rogueSpellInfo.nightbladeDebuff, op = HoJ.OP_LESS },
            { aType = HoJ.TYPE_RESULT, aConst = 8, bType = HoJ.TYPE_RESULT, bConst = 9, op = HoJ.OP_OR },
            { aType = HoJ.TYPE_RESULT, aConst = 5, bType = HoJ.TYPE_RESULT, bConst = 10, op = HoJ.OP_AND },
        }
    },
    {
        -- eviscerate at 4+ combo points if not talented for Deeper Strategem
        spellInfo = rogueSpellInfo.eviscerate,
        _debugName = "eviscerate",
        tests = {
            { aType = HoJ.TYPE_TALENT_SELECTED, aConst = rogueSpellInfo.deeperStrategem, bType = HoJ.TYPE_CONSTANT, bConst = false, op = HoJ.OP_EQUAL },
            { aType = HoJ.TYPE_COMBO_POINTS, aConst = 1, bType = HoJ.TYPE_CONSTANT, bConst = 4, op = HoJ.OP_GREATER_EQUAL },
            { aType = HoJ.TYPE_RESULT, aConst = 1, bType = HoJ.TYPE_RESULT, bConst = 2, op = HoJ.OP_AND },
            { aType = HoJ.TYPE_TARGET_RANGE, aConst = rogueSpellInfo.eviscerate.range, bType = HoJ.TYPE_CONSTANT, bConst = true, op = HoJ.OP_EQUAL },
            { aType = HoJ.TYPE_RESULT, aConst = 3, bType = HoJ.TYPE_RESULT, bConst = 4, op = HoJ.OP_AND },
            { aType = HoJ.TYPE_PLAYER_BUFF_ACTIVE, aConst = rogueSpellInfo.eviscerate.id, bType = HoJ.TYPE_CONSTANT, bConst = true, op = HoJ.OP_EQUAL },
            { aType = HoJ.TYPE_TARGET_PLAYER_DEBUFF_REMAINING, aConst = rogueSpellInfo.nightbladeDebuff.id, bType = HoJ.TYPE_CONSTANT, bConst = 0.0, op = HoJ.OP_GREATER },
            { aType = HoJ.TYPE_RESULT, aConst = 6, bType = HoJ.TYPE_RESULT, bConst = 7, op = HoJ.OP_OR },
            { aType = HoJ.TYPE_COMBAT_TARGET_TIME_TO_LIVE, aConst = 2, bType = HoJ.TYPE_BUFF_DURATION_FOR_COMBO_POINTS, bConst = rogueSpellInfo.nightbladeDebuff, op = HoJ.OP_LESS },
            { aType = HoJ.TYPE_RESULT, aConst = 8, bType = HoJ.TYPE_RESULT, bConst = 9, op = HoJ.OP_OR },
            { aType = HoJ.TYPE_RESULT, aConst = 5, bType = HoJ.TYPE_RESULT, bConst = 10, op = HoJ.OP_AND },
        }
    },
    {
        spellInfo = rogueSpellInfo.shadowstrike,
        _debugName = "shadowstrike",
        tests = {
            { aType = HoJ.TYPE_PLAYER_IS_STEALTHED, aConst = 1, bType = HoJ.TYPE_CONSTANT, bConst = true, op = HoJ.OP_EQUAL },
            { aType = HoJ.TYPE_PLAYER_BUFF_ACTIVE, aConst = rogueSpellInfo.shadowDanceBuff.id, bType = HoJ.TYPE_CONSTANT, bConst = true, op = HoJ.OP_EQUAL },
            { aType = HoJ.TYPE_RESULT, aConst = 1, bType = HoJ.TYPE_RESULT, bConst = 2, op = HoJ.OP_OR },
            { aType = HoJ.TYPE_TARGET_RANGE, aConst = rogueSpellInfo.shadowstrike.range, bType = HoJ.TYPE_CONSTANT, bConst = true, op = HoJ.OP_EQUAL },
            { aType = HoJ.TYPE_NUM_MELEE_ATTACKERS, aConst = 1, bType = HoJ.TYPE_CONSTANT, bConst = 3, op = HoJ.OP_LESS },
            { aType = HoJ.TYPE_RESULT, aConst = 4, bType = HoJ.TYPE_RESULT, bConst = 5, op = HoJ.OP_AND },
            { aType = HoJ.TYPE_RESULT, aConst = 3, bType = HoJ.TYPE_RESULT, bConst = 6, op = HoJ.OP_AND },
        }
    },
    {
        -- shuriken storm as long as we will have > 50 energy when symbols of death comes off cooldown
        spellInfo = rogueSpellInfo.shurikenStorm,
        _debugName = "shurikenStorm",
        tests = {
            { 
                aType = HoJ.TYPE_POWER_AFTER_CD_MINUS_SKILL_POWER,
                aConst = rogueSpellInfo.symbolsOfDeath.id,
                aConst2 = rogueSpellInfo.shurikenStorm,
                bType = HoJ.TYPE_CONSTANT,
                bConst = 50,
                op = HoJ.OP_GREATER_EQUAL
            },
            { aType = HoJ.TYPE_TARGET_RANGE, aConst = rogueSpellInfo.shurikenStorm.range, bType = HoJ.TYPE_CONSTANT, bConst = true, op = HoJ.OP_EQUAL },
            { aType = HoJ.TYPE_RESULT, aConst = 1, bType = HoJ.TYPE_RESULT, bConst = 2, op = HoJ.OP_AND },
        }
    },
}

local subtletyMitigationRules = {
    _name = "subtletyMitigationRules",
    {
        spellInfo = rogueSpellInfo.evasion,
        -- _debugName = "evasion",
        tests = {
            { aType = HoJ.TYPE_PLAYER_TIME_TO_LIVE, aConst = 0, bType = HoJ.TYPE_CONSTANT, bConst = 10.0, op = HoJ.OP_LESS },
            { aType = HoJ.TYPE_NET_INCOMING_DPS_CHR, aConst = 5.0, bType = HoJ.TYPE_PLAYER_HEALTH_RATIO, bConst = 0 , op = HoJ.OP_GREATER_EQUAL },
            { aType = HoJ.TYPE_RESULT, aConst = 1, bType = HoJ.TYPE_RESULT, bConst = 2, op = HoJ.OP_OR },
        }
    },
    -- TODO: feint when taking AoE damage
    {
        spellInfo = rogueSpellInfo.crimsonVial,
        _debugName = "crimsonVial",
        tests = {
            { aType = HoJ.TYPE_PLAYER_HEALTH_RATIO, aConst = 1, bType = HoJ.TYPE_MAJOR_HEAL_HEALTH_RATIO, bConst = 0, op = HoJ.OP_LESS_EQUAL },
            { aType = HoJ.TYPE_NET_INCOMING_DPS_CHR, aConst = 10.0, bType = HoJ.TYPE_PLAYER_HEALTH_RATIO, bConst = 0 , op = HoJ.OP_GREATER_EQUAL },
            { aType = HoJ.TYPE_RESULT, aConst = 1, bType = HoJ.TYPE_RESULT, bConst = 2, op = HoJ.OP_OR },
            { aType = HoJ.TYPE_SKILL_POWER_COST, aConst = rogueSpellInfo.kidneyShot, bType = HoJ.TYPE_POWER, bConst = 0, op = HoJ.OP_LESS_EQUAL },
            { aType = HoJ.TYPE_RESULT, aConst = 3, bType = HoJ.TYPE_RESULT, bConst = 4, op = HoJ.OP_AND },
            { aType = HoJ.TYPE_RESULT, aConst = 3, bType = HoJ.TYPE_RESULT, bConst = 5, op = HoJ.OP_AND },
        }
    },
}

local subtletyInterruptRules = {
    _name = "subtletyInterruptRules",
    {
        spellInfo = rogueSpellInfo.kick,
        _debugName = "kick",
        tests = {
            { aType = HoJ.TYPE_TARGET_RANGE, aConst = rogueSpellInfo.kick.range, bType = HoJ.TYPE_CONSTANT, bConst = true, op = HoJ.OP_EQUAL },
            { aType = HoJ.TYPE_TARGET_CAN_BE_INTERRUPTED, aConst = 0, bType = HoJ.TYPE_CONSTANT, bConst = true, op = HoJ.OP_EQUAL },
            { aType = HoJ.TYPE_RESULT, aConst = 1, bType = HoJ.TYPE_RESULT, bConst = 2, op = HoJ.OP_AND },
        }
    },
    {
        spellInfo = rogueSpellInfo.kidneyShot,
        --_debugName = "kidneyShot",
        tests = {
            { aType = HoJ.TYPE_TARGET_RANGE, aConst = rogueSpellInfo.kidneyShot.range, bType = HoJ.TYPE_CONSTANT, bConst = true, op = HoJ.OP_EQUAL },
            { aType = HoJ.TYPE_TARGET_CAN_BE_INTERRUPTED, aConst = 0, bType = HoJ.TYPE_CONSTANT, bConst = true, op = HoJ.OP_EQUAL },
            { aType = HoJ.TYPE_RESULT, aConst = 1, bType = HoJ.TYPE_RESULT, bConst = 2, op = HoJ.OP_AND },
            { aType = HoJ.TYPE_TARGET_IS_BOSS, aConst = 0, bType = HoJ.TYPE_CONSTANT, bConst = false, op = HoJ.OP_EQUAL },
            { aType = HoJ.TYPE_SKILL_POWER_COST, aConst = rogueSpellInfo.kidneyShot, bType = HoJ.TYPE_POWER, bConst = 0, op = HoJ.OP_LESS_EQUAL },
            { aType = HoJ.TYPE_RESULT, aConst = 4, bType = HoJ.TYPE_RESULT, bConst = 5, op = HoJ.OP_AND },
            { aType = HoJ.TYPE_RESULT, aConst = 3, bType = HoJ.TYPE_RESULT, bConst = 6, op = HoJ.OP_AND },
        }
    },
}

local subtletyDpsCooldownRules = {
    _name = "subtletyDpsCooldownRules",
    {
        -- show shadow blands in the dps cooldowns if we're in multitarget, since it's discresionary
        spellInfo = rogueSpellInfo.shadowBlades,
        _debugName = "shadowBlades",
        tests = {
            { aType = HoJ.TYPE_NUM_MELEE_ATTACKERS, aConst = 1, bType = HoJ.TYPE_CONSTANT, bConst = 2, op = HoJ.OP_GREATER_EQUAL },
        }
    },
}

local subtletyCooldownSortPriorities = {
    _name = "subtletyCooldownSortPriorities",
    [rogueSpellInfo.shadowBlades.id] = 6,
    [rogueSpellInfo.evasion.id] = 5,
    [rogueSpellInfo.crimsonVial.id] = 4,
    [rogueSpellInfo.kick.id] = 3,
    [rogueSpellInfo.kidneyShot.id] = 2,
    [rogueSpellInfo.feint.id] = 1,
}

local subtletyActiveBuffList = { 
    [rogueSpellInfo.shadowDanceBuff.id] = { isActive = false, start = 0, duration = 0, timeRemaining = 0, stacks = 0 },
    [rogueSpellInfo.symbolsOfDeathBuff.id] = { isActive = false, start = 0, duration = 0, timeRemaining = 0, stacks = 0 },
}

-- this is the list of buffs that we want to show on the buffs bar
-- in general we just show buffs that proc randomly here, or buffs with a stack count that needs tracking
local subtletyProcdBuffList = {
    { _spellInfo = rogueSpellInfo.shadowDanceBuff },
    { _spellInfo = rogueSpellInfo.symbolsOfDeathBuff },
}

local subtletyActiveBlessingList = {
--[[
    [rogueSpellInfo.battleShout.id] = {
        spellInfo = rogueSpellInfo.battleShout,
        targetName = nil,
        start = 0.0,
        duration = 0.0,
        updateIndex = -1
    }
--]]
}

local subtletyBuffSortPriorities = {
    _name = "subtletyBuffSortPriorities",
    [rogueSpellInfo.shadowDanceBuff.id] = 2,
    [rogueSpellInfo.symbolsOfDeathBuff.id] = 1,
}

HoJ[HoJ.CLASS_NAME_ROGUE] = {
    _spellInfos = HoJ._rogueSpellInfos,
    _className = HoJ.CLASS_NAME_ROGUE,
    _defaultSpecId = HoJ.SPEC_ROGUE_ASSASSINATION,
    _powerColor = { r = 1.0, g = 1.0, b = 0.0, a = 1.0 },
    _usesComboPoints = true,
    [HoJ.SPEC_NAME_ROGUE_SUBTLETY] = {
        -- properties
        _name = HoJ._text["Subtlety"],
        _spec = HoJ.SPEC_ROGUE_SUBTLETY,
        _role = HoJ.ROLE_DPS,
        _enabled = false,
        _multiTargetRange = 10,
        _minorHealHealthRatio = 0.9,
        _majorHealHealthRatio = 0.6,
        _defaultPowerPerSec = 10, -- testing shows around 10 pps
        _autoPowerGen = true,
        _powerType = 3, -- 3 = energy
        _powerGainStep = 1,
        _powerMax = 100,
        _usesPowerPooling = true,
        -- functions
        _updateFn = HoJ.Rogue_Update,
        _updateRenderFn = nil,
        _resetFn = nil,
        _selectSingleOrMultiTargetFn = HoJ.Rogue_SelectSingleOrMultitarget,
        _nextSkillSelectFn = HoJ.NextSkillOffCooldown,
        _checkBlessingsFn = nil,
        -- tables
        _meleeAoESpells = nil,
        _activeBuffList = subtletyActiveBuffList,
        _procdBuffList = subtletyProcdBuffList,
        _activeBlessings = subtletyActiveBlessingList,
        _rotationRules = { _singleTarget = subtletyRotationRules, _multiTarget = subtletyMultiTargetRotationRules },
        _interruptRules = { _singleTarget = subtletyInterruptRules, _multiTarget = subtletyInterruptRules },
        _mitigationRules = { _singleTarget = subtletyMitigationRules, _multiTarget = subtletyMitigationRules },
        _dpsCooldownRules = { _singleTarget = subtletyDpsCooldownRules, _multiTarget = subtletyDpsCooldownRules },
        _cooldownSortPriorities = subtletyCooldownSortPriorities,
        _buffSortPriorities = subtletyBuffSortPriorities,
        _friendlyRangeSpells = {
            [HoJ.RANGE_MELEE] = nil,
            [HoJ.RANGE_8_YARDS] = nil,
            [HoJ.RANGE_10_YARDS] = nil,
            [HoJ.RANGE_12_YARDS] = nil,
            [HoJ.RANGE_20_YARDS] = nil,
            [HoJ.RANGE_25_YARDS] = nil,
            [HoJ.RANGE_30_YARDS] = nil,
            [HoJ.RANGE_40_YARDS] = nil,
            [HoJ.RANGE_100_YARDS] = {
                _spellInfo = rogueSpellInfo.tricksOfTheTrade,
            },
        },
        _enemyRangeSpells = {
            [HoJ.RANGE_MELEE] = {
                _spellInfo = rogueSpellInfo.eviscerate
            },
            [HoJ.RANGE_8_YARDS] = nil,
            [HoJ.RANGE_10_YARDS] = nil,
            [HoJ.RANGE_12_YARDS] = nil,
            [HoJ.RANGE_15_YARDS] = {
                _spellInfo = rogueSpellInfo.blind,
            },
            [HoJ.RANGE_20_YARDS] = {
                _spellInfo = rogueSpellInfo.blind,
            },
            [HoJ.RANGE_25_YARDS] = {
                _spellInfo = rogueSpellInfo.shadowstep,
            },
            [HoJ.RANGE_30_YARDS] = {
                _spellInfo = rogueSpellInfo.shadowstep
            },
            [HoJ.RANGE_40_YARDS] = {
                _spellInfo = rogueSpellInfo.shadowstep
            },
            [HoJ.RANGE_100_YARDS] = {
                _spellInfo = rogueSpellInfo.shadowstep
            },
        },
    },
    [HoJ.SPEC_NAME_ROGUE_ASSASSINATION] = {
        -- properties
        _name = HoJ._text["Assassination"],
        _spec = HoJ.SPEC_ROGUE_ASSASSINATION,
        _role = HoJ.ROLE_DPS,
        _enabled = false,
        _multiTargetRange = 10,
        _minorHealHealthRatio = 0.9,
        _majorHealHealthRatio = 0.6,
        _powerType = 3, -- 3 = energy
    },
    [HoJ.SPEC_NAME_ROGUE_OUTLAW] = {
        -- properties
        _name = HoJ._text["Outlaw"],
        _spec = HoJ.SPEC_ROGUE_OUTLAW,
        _role = HoJ.ROLE_DPS,
        _enabled = false,
        _multiTargetRange = 10,
        _minorHealHealthRatio = 0.9,
        _majorHealHealthRatio = 0.6,
        _powerType = 3, -- 3 = energy
    },
}

-- -----------------------------------------------------------------------------
-- HoJ:LoadRogue
-- -----------------------------------------------------------------------------
function HoJ:LoadRogue()
    if ( HoJ._classId ~= HoJ.CLASS_ID_ROGUE ) then
        HoJ:Error( "LoadRogue called but _classId is not rogue!" )
    end

    HoJ.SpellInfo = HoJ._rogueSpellInfos
end
